"""
Azure Container Instances deployment.
"""

from __future__ import annotations

__all__ = ["AzureContainerInstances"]

from typing import Any

from azure.core.exceptions import ResourceNotFoundError
from azure.mgmt.containerinstance import ContainerInstanceManagementClient
from azure.mgmt.containerinstance.aio import (
    ContainerInstanceManagementClient as AContainerInstanceManagementClient,
)
from azure.mgmt.containerinstance.models import Container as ACIContainer
from azure.mgmt.containerinstance.models import (
    ContainerExec,
    ContainerGroup,
    ContainerGroupDiagnostics,
    ContainerGroupIpAddressType,
    ContainerGroupRestartPolicy,
    ContainerGroupSku,
    ContainerHttpGet,
    ContainerNetworkProtocol,
    ContainerPort,
    ContainerProbe,
    EnvironmentVariable,
)
from azure.mgmt.containerinstance.models import GpuResource as ACIGpuResource
from azure.mgmt.containerinstance.models import GpuSku
from azure.mgmt.containerinstance.models import HttpHeader as ACIHttpHeader
from azure.mgmt.containerinstance.models import (
    ImageRegistryCredential,
    InitContainerDefinition,
    IpAddress,
    LogAnalytics,
    OperatingSystemTypes,
)
from azure.mgmt.containerinstance.models import Port as ACIPort
from azure.mgmt.containerinstance.models import (
    ResourceLimits as ACIResourceLimits,
)
from azure.mgmt.containerinstance.models import (
    ResourceRequests as ACIResourceRequests,
)
from azure.mgmt.containerinstance.models import (
    ResourceRequirements as ACIResourceRequirements,
)
from azure.mgmt.containerinstance.models import (
    SecurityContextCapabilitiesDefinition,
    SecurityContextDefinition,
)
from azure.mgmt.containerinstance.models import Volume as ACIVolume
from azure.mgmt.containerinstance.models import VolumeMount as ACIVolumeMount

from x8._common.azure_provider import AzureProvider
from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Context, OperationParser, Response, RunContext
from x8.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import Expression

from .._feature import ContainerDeploymentFeature
from .._helper import amap_images, map_images, requires_container_registry
from .._models import (
    Container,
    ContainerRegistryCredentials,
    EnvVar,
    ExecAction,
    GPUResource,
    HTTPGetAction,
    HTTPHeader,
    Ingress,
    Port,
    Probe,
    ProbeSet,
    ResourceLimits,
    ResourceRequests,
    ResourceRequirements,
    Revision,
    SecurityContext,
    ServiceDefinition,
    ServiceItem,
    ServiceOverlay,
    TrafficAllocation,
    Volume,
    VolumeMount,
)
from ._base import BaseContainerDeploymentProvider


class AzureContainerInstances(AzureProvider, BaseContainerDeploymentProvider):
    subscription_id: str | None
    resource_group: str | None
    location: str
    container_group_name: str | None
    container_registry_credentials: list[ContainerRegistryCredentials] | None
    os_type: str
    sku: str | None
    log_analytics_workspace_id: str | None
    log_analytics_workspace_key: str | None
    nparams: dict[str, Any]

    _client: ContainerInstanceManagementClient
    _aclient: AContainerInstanceManagementClient
    _credential: Any
    _acredential: Any
    _init: bool = False
    _ainit: bool = False
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _cached_subscription_id: str | None = None

    def __init__(
        self,
        subscription_id: str | None = None,
        resource_group: str | None = None,
        location: str = "westus2",
        container_group_name: str | None = None,
        service: ServiceDefinition | None = None,
        service_override: ServiceOverlay | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        container_registry_credentials: (
            list[ContainerRegistryCredentials] | None
        ) = None,
        os_type: str = "Linux",
        sku: str | None = None,
        log_analytics_workspace_id: str | None = None,
        log_analytics_workspace_key: str | None = None,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            subscription_id: Azure subscription ID.
            resource_group: The resource group to deploy to.
            location: Azure region for the container instance.
            container_group_name: The name of the container group to deploy.
            service: Service definition to deploy.
            service_override: Service override definition to apply.
            containerizer: Containerizer instance for building images.
            container_registry: Container registry instance for pushing images.
            container_registry_credentials:
                Credentials for the container registry.
            os_type: Operating system type for the container.
            sku: SKU for the container group.
            credential_type:
                Type of Azure credential to use. Options are:
                - default: DefaultAzureCredential
                - client_secret: ClientSecretCredential
                - certificate: CertificateCredential
                - azure_cli: AzureCliCredential
                - shared_token_cache: SharedTokenCacheCredential
                - managed_identity: ManagedIdentityCredential
            tenant_id: Tenant ID for Azure authentication.
            client_id: Client ID for Azure authentication.
            client_secret: Client secret for Azure authentication.
            certificate_path: Path to certificate file for authentication.
            nparams: Native params to Azure client.
        """
        self.subscription_id = subscription_id
        self.resource_group = resource_group
        self.location = location
        self.container_group_name = container_group_name
        self.service = service
        self.overlay = service_override
        self.containerizer = containerizer
        self.container_registry = container_registry
        self.container_registry_credentials = container_registry_credentials
        self.os_type = os_type
        self.sku = sku
        self.log_analytics_workspace_id = log_analytics_workspace_id
        self.log_analytics_workspace_key = log_analytics_workspace_key
        self.nparams = nparams

        self._init = False
        self._ainit = False
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()
        self._credential = None
        self._acredential = None
        self._cached_subscription_id = None

        AzureProvider.__init__(
            self,
            credential_type=credential_type,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            certificate_path=certificate_path,
            **kwargs,
        )

    def __supports__(self, feature):
        return feature not in [
            ContainerDeploymentFeature.MULTIPLE_REVISIONS,
            ContainerDeploymentFeature.REVISION_DELETE,
        ]

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        self._credential = self._get_credential()
        subscription_id = self._get_subscription_id()
        self._client = ContainerInstanceManagementClient(
            credential=self._credential,
            subscription_id=subscription_id,
            **self.nparams,
        )
        self._init = True

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._ainit:
            return

        self._acredential = self._aget_credential()
        subscription_id = self._get_subscription_id()
        self._aclient = AContainerInstanceManagementClient(
            credential=self._acredential,
            subscription_id=subscription_id,
            **self.nparams,
        )
        self._ainit = True

    def _init_credentials(self):
        if not self._credential:
            self._credential = self._get_credential()

    def _ensure_container_registry(
        self, service: ServiceDefinition
    ) -> ContainerRegistry | None:
        container_registry = self._get_container_registry()
        if container_registry:
            return container_registry
        if not requires_container_registry(service):
            return None

        from x8.compute.container_registry.providers.azure_container_registry import (  # noqa
            AzureContainerRegistry,
        )

        acr_name = f"{self._get_container_group_name(service)}-acr".replace(
            "_", ""
        ).replace("-", "")
        acr = AzureContainerRegistry(
            name=acr_name,
            credential_type=self.credential_type,
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
            certificate_path=self.certificate_path,
        )
        acr.create_resource(
            resource_group=self._get_resource_group_name(
                self._get_container_group_name(service)
            ),
            location=self.location,
            subscription_id=self.subscription_id,
        )
        print(f"Created container registry: {acr_name}")
        return ContainerRegistry(__provider__=acr)

    def _check_delete_container_registry(self, name: str) -> None:
        container_registry = self._get_container_registry()
        if container_registry:
            return None

        from x8.compute.container_registry.providers.azure_container_registry import (  # noqa
            AzureContainerRegistry,
        )

        acr_name = f"{name}-acr".replace("_", "").replace("-", "")
        acr = AzureContainerRegistry(
            name=acr_name,
            credential_type=self.credential_type,
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
            certificate_path=self.certificate_path,
        )
        try:
            acr.get_resource(
                resource_group=self._get_resource_group_name(name),
                name=acr_name,
                subscription_id=self.subscription_id,
            )
            acr.delete_resource(
                resource_group=self._get_resource_group_name(name),
                subscription_id=self.subscription_id,
            )
            print(f"Deleted container registry: {acr_name}")
        except NotFoundError:
            pass

    def create_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        self.__setup__()
        where_exists = OperationParser.parse_where_exists(where)
        service_def = self._normalize_service_definition(service)
        container_group_name = self._get_container_group_name(service_def)
        subscription_id = self._get_subscription_id()
        resource_group_name = self._get_resource_group_name(
            container_group_name=container_group_name
        )
        try:
            existing_service = self._client.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=container_group_name,
            )
            if where_exists is False:
                raise PreconditionFailedError(
                    f"Service {container_group_name} already exists."
                )
            self._ensure_resource_group(
                resource_group=resource_group_name,
                location=self.location,
                subscription_id=subscription_id,
            )
            images = map_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            image_registry_credentials = self._get_image_registry_credentials(
                subscription_id=subscription_id,
                images=images,
            )
            updated_service = self._op_converter.update_service(
                existing_service=existing_service,
                service=service_def,
                images=images,
                container_group_name=container_group_name,
                location=self.location,
                image_registry_credentials=image_registry_credentials,
                diagnostics=self._get_diagnostics(),
                os_type=self.os_type,
                sku=self.sku,
            )

            operation = self._client.container_groups.begin_create_or_update(
                resource_group_name=resource_group_name,
                container_group_name=container_group_name,
                container_group=updated_service,
            )
            response = operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service updated successfully: {result.uri}")
        except ResourceNotFoundError:
            if where_exists is True:
                raise PreconditionFailedError(
                    f"Service {container_group_name} not found."
                )
            self._ensure_resource_group(
                resource_group=resource_group_name,
                location=self.location,
                subscription_id=subscription_id,
            )
            images = map_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            image_registry_credentials = self._get_image_registry_credentials(
                subscription_id=subscription_id,
                images=images,
            )
            new_service = self._op_converter.convert_service(
                service=service_def,
                images=images,
                container_group_name=container_group_name,
                location=self.location,
                image_registry_credentials=image_registry_credentials,
                diagnostics=self._get_diagnostics(),
                os_type=self.os_type,
                sku=self.sku,
            )
            operation = self._client.container_groups.begin_create_or_update(
                resource_group_name=resource_group_name,
                container_group_name=container_group_name,
                container_group=new_service,
            )
            response = operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service created successfully: {result.uri}")
        return Response(result=result)

    async def acreate_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        await self.__asetup__()
        where_exists = OperationParser.parse_where_exists(where)
        service_def = self._normalize_service_definition(service)
        container_group_name = self._get_container_group_name(service_def)
        subscription_id = self._get_subscription_id()
        resource_group_name = self._get_resource_group_name(
            container_group_name=container_group_name
        )
        try:
            existing_service = await self._aclient.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=container_group_name,
            )
            if where_exists is False:
                raise PreconditionFailedError(
                    f"Service {container_group_name} already exists."
                )
            images = await amap_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            image_registry_credentials = self._get_image_registry_credentials(
                subscription_id=subscription_id, images=images
            )
            self._ensure_resource_group(
                resource_group=resource_group_name,
                location=self.location,
                subscription_id=subscription_id,
            )
            updated_service = self._op_converter.update_service(
                existing_service=existing_service,
                service=service_def,
                images=images,
                container_group_name=container_group_name,
                location=self.location,
                image_registry_credentials=image_registry_credentials,
                diagnostics=self._get_diagnostics(),
                os_type=self.os_type,
                sku=self.sku,
            )
            operation = (
                await self._aclient.container_groups.begin_create_or_update(
                    resource_group_name=resource_group_name,
                    container_group_name=container_group_name,
                    container_group=updated_service,
                )
            )
            response = await operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service updated successfully: {result.uri}")
        except ResourceNotFoundError:
            if where_exists is True:
                raise PreconditionFailedError(
                    f"Service {container_group_name} not found."
                )
            images = await amap_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            image_registry_credentials = self._get_image_registry_credentials(
                subscription_id=subscription_id, images=images
            )
            new_service = self._op_converter.convert_service(
                service=service_def,
                images=images,
                container_group_name=container_group_name,
                location=self.location,
                image_registry_credentials=image_registry_credentials,
                diagnostics=self._get_diagnostics(),
                os_type=self.os_type,
                sku=self.sku,
            )
            self._ensure_resource_group(
                resource_group=resource_group_name,
                location=self.location,
                subscription_id=subscription_id,
            )
            operation = (
                await self._aclient.container_groups.begin_create_or_update(
                    resource_group_name=resource_group_name,
                    container_group_name=container_group_name,
                    container_group=new_service,
                )
            )
            response = await operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service created successfully: {result.uri}")
        return Response(result=result)

    def get_service(self, name: str, **kwargs: Any) -> Response[ServiceItem]:
        self.__setup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            response = self._client.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            result = self._result_converter.convert_service_item(response)
            return Response(result=result)
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    async def aget_service(
        self, name: str, **kwargs: Any
    ) -> Response[ServiceItem]:
        await self.__asetup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            response = await self._aclient.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            result = self._result_converter.convert_service_item(response)
            return Response(result=result)
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    def delete_service(self, name: str, **kwargs: Any) -> Response[None]:
        self.__setup__()
        subscription_id = self._get_subscription_id()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            self._client.container_groups.begin_delete(
                resource_group_name=resource_group_name,
                container_group_name=name,
            ).result()
            print(f"Service {name} deleted successfully.")
            self._check_delete_container_registry(name)
            if not self.resource_group:
                self._delete_resource_group_if_empty(
                    subscription_id=subscription_id,
                    resource_group=resource_group_name,
                )
            return Response(result=None)
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    async def adelete_service(
        self, name: str, **kwargs: Any
    ) -> Response[None]:
        await self.__asetup__()
        subscription_id = self._get_subscription_id()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            operation = await self._aclient.container_groups.begin_delete(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            await operation.result()
            print(f"Service {name} deleted successfully.")
            self._check_delete_container_registry(name)
            if not self.resource_group:
                self._delete_resource_group_if_empty(
                    subscription_id=subscription_id,
                    resource_group=resource_group_name,
                )
            return Response(result=None)
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    def list_services(self, **kwargs: Any) -> Response[list[ServiceItem]]:
        self.__setup__()
        response = self._client.container_groups.list()
        result = self._result_converter.convert_services(
            services=list(response)
        )
        return Response(result=result)

    async def alist_services(
        self, **kwargs: Any
    ) -> Response[list[ServiceItem]]:
        await self.__asetup__()
        response = self._aclient.container_groups.list()
        result = self._result_converter.convert_services(
            services=[service async for service in response]
        )
        return Response(result=result)

    def list_revisions(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[list[Revision]]:
        self.__setup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            response = self._client.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            result = self._result_converter.convert_service_item(response)
            revision = Revision(
                name="latest",
                traffic=100,
                containers=result.service.containers if result.service else [],
                volumes=result.service.volumes if result.service else [],
                created_time=None,
                active=True,
                status="running",
            )
            return Response(result=[revision])
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    async def alist_revisions(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[list[Revision]]:
        self.__setup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            response = await self._aclient.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            result = self._result_converter.convert_service_item(response)
            revision = Revision(
                name="latest",
                traffic=100,
                containers=result.service.containers if result.service else [],
                volumes=result.service.volumes if result.service else [],
                created_time=None,
                active=True,
                status="running",
            )
            return Response(result=[revision])
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    def get_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[Revision]:
        if revision != "latest":
            raise NotFoundError("Only 'latest' revision is supported.")
        self.__setup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            response = self._client.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            result = self._result_converter.convert_service_item(response)
            return Response(
                result=Revision(
                    name="latest",
                    traffic=100,
                    containers=(
                        result.service.containers if result.service else []
                    ),
                    volumes=result.service.volumes if result.service else [],
                    created_time=None,
                    active=True,
                    status="running",
                )
            )
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    async def aget_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[Revision]:
        if revision != "latest":
            raise NotFoundError("Only 'latest' revision is supported.")
        await self.__asetup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        try:
            response = await self._aclient.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
            result = self._result_converter.convert_service_item(response)
            return Response(
                result=Revision(
                    name="latest",
                    traffic=100,
                    containers=(
                        result.service.containers if result.service else []
                    ),
                    volumes=result.service.volumes if result.service else [],
                    created_time=None,
                    active=True,
                    status="running",
                )
            )
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )

    def delete_revision(self, name: str, revision: str) -> Response[None]:
        self.__setup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        if revision != "latest":
            raise NotFoundError("Only 'latest' revision is supported.")
        try:
            self._client.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )
        raise PreconditionFailedError(
            (
                f"Cannot delete active revision {revision} "
                f"of service {name}."
            )
        )

    async def adelete_revision(
        self, name: str, revision: str
    ) -> Response[None]:
        await self.__asetup__()
        resource_group_name = self._get_resource_group_name(
            container_group_name=name
        )
        if revision != "latest":
            raise NotFoundError("Only 'latest' revision is supported.")
        try:
            await self._aclient.container_groups.get(
                resource_group_name=resource_group_name,
                container_group_name=name,
            )
        except ResourceNotFoundError:
            raise NotFoundError(
                (
                    f"Service {name} not found in resource group "
                    f"{resource_group_name}."
                )
            )
        raise PreconditionFailedError(
            (
                f"Cannot delete active revision {revision} "
                f"of service {name}."
            )
        )

    def update_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
    ) -> Response[ServiceItem]:
        return self.get_service(name)

    async def aupdate_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
    ) -> Response[ServiceItem]:
        return await self.aget_service(name)

    def close(self) -> None:
        self._init = False

    async def aclose(self) -> None:
        if hasattr(self, "_aclient") and self._aclient:
            await self._aclient.close()
        if hasattr(self, "_acredential") and self._acredential:
            await self._acredential.close()
        self._ainit = False

    def _get_subscription_id(self) -> str:
        if self.subscription_id:
            return self.subscription_id
        if self._cached_subscription_id:
            return self._cached_subscription_id
        self._cached_subscription_id = self._get_default_subscription_id()
        return self._cached_subscription_id

    def _get_default_subscription_id(self) -> str:
        from azure.mgmt.resource import SubscriptionClient

        self._init_credentials()
        sub_client = SubscriptionClient(credential=self._credential)
        subs = sub_client.subscriptions.list()
        if not subs:
            raise BadRequestError(
                "No subscriptions found for the given credentials."
            )
        sub = next(sub_client.subscriptions.list())
        return sub.subscription_id

    def _get_diagnostics(self) -> ContainerGroupDiagnostics | None:
        if (
            self.log_analytics_workspace_id
            and self.log_analytics_workspace_key
        ):
            return ContainerGroupDiagnostics(
                log_analytics=LogAnalytics(
                    workspace_id=self.log_analytics_workspace_id,
                    workspace_key=self.log_analytics_workspace_key,
                )
            )
        return None

    def _get_container_group_name(self, service: ServiceDefinition) -> str:
        if self.container_group_name:
            return self.container_group_name
        if service.name:
            return service.name
        raise BadRequestError(
            "Container group name must be provided or defined in the service."
        )

    def _get_resource_group_name(self, container_group_name: str) -> str:
        """Get the resource group name for the Container App."""
        if self.resource_group:
            return self.resource_group
        return f"{container_group_name}-rg"

    def _get_image_registry_credentials(
        self, subscription_id: str, images: list[str]
    ) -> list[ImageRegistryCredential]:
        """Get registry credentials for ACR authentication."""
        credentials: list[ImageRegistryCredential] = []
        if self.container_registry_credentials is None:
            container_registry_credentials: list[
                ContainerRegistryCredentials
            ] = []
            hash_set = set()
            for image_uri in images:
                if ".azurecr.io" in image_uri:
                    acr_name = image_uri.split(".")[0]
                    server = f"{acr_name}.azurecr.io"
                    if server not in hash_set:
                        container_registry_credentials.append(
                            ContainerRegistryCredentials(
                                server=server, auth_type="basic"
                            )
                        )
                        hash_set.add(server)
        else:
            container_registry_credentials = (
                self.container_registry_credentials
            )

        for cred in container_registry_credentials:
            if cred.auth_type == "managed":
                continue
            elif cred.auth_type == "basic":
                if cred.username and cred.password:
                    credentials.append(
                        ImageRegistryCredential(
                            server=cred.server,
                            username=cred.username,
                            password=cred.password,
                        )
                    )
                elif ".azurecr.io" in cred.server:
                    try:
                        from azure.mgmt.containerregistry import (
                            ContainerRegistryManagementClient,
                        )

                        self._init_credentials()

                        acr_client = ContainerRegistryManagementClient(
                            credential=self._credential,
                            subscription_id=subscription_id,
                        )

                        target_server = cred.server.lower()
                        found = None
                        for reg in acr_client.registries.list():
                            # reg.login_server looks like 'myacr.azurecr.io'
                            if reg.login_server == target_server:
                                found = reg
                                break

                        if not found:
                            print(
                                f"Warning: ACR with login server '{cred.server}' not found in subscription {subscription_id}."  # noqa
                            )
                            continue

                        # Parse resource group from resource ID:
                        # /subscriptions/{sub}/resourceGroups/{rg}/providers/Microsoft.ContainerRegistry/registries/{name}
                        reg_id = found.id
                        parts = reg_id.split("/")
                        try:
                            rg_name = parts[parts.index("resourceGroups") + 1]
                        except Exception:
                            print(
                                f"Warning: Could not parse resource group from ACR id: {reg_id}"  # noqa
                            )
                            continue

                        reg_name = found.name
                        if not reg_name:
                            print(
                                "Warning: ACR resource missing name; cannot list credentials."  # noqa
                            )
                            continue

                        # Requires ACR Admin user to be enabled.
                        cred_result = acr_client.registries.list_credentials(
                            resource_group_name=rg_name,
                            registry_name=reg_name,
                        )
                        user = cred_result.username
                        pwds = cred_result.passwords or []
                        pwd_value = (
                            pwds[0].value if pwds and pwds[0].value else None
                        )

                        if user and pwd_value:
                            credentials.append(
                                ImageRegistryCredential(
                                    server=cred.server,
                                    username=user,
                                    password=pwd_value,
                                )
                            )
                        else:
                            print(
                                (
                                    "Warning: ACR admin user may be disabled or returned no passwords; "  # noqa
                                    "consider using auth_type='managed' or provide explicit basic credentials."  # noqa
                                )
                            )
                    except Exception as e:
                        print(
                            f"Warning: Could not retrieve ACR credentials: {e}"
                        )
        return credentials


class OperationConverter:
    def update_service(
        self,
        existing_service: ContainerGroup,
        service: ServiceDefinition,
        images: list[str],
        container_group_name: str,
        location: str,
        image_registry_credentials: (
            list[ImageRegistryCredential] | None
        ) = None,
        diagnostics: ContainerGroupDiagnostics | None = None,
        os_type: str | OperatingSystemTypes = OperatingSystemTypes.LINUX,
        sku: str | ContainerGroupSku | None = None,
    ) -> ContainerGroup:
        new_service = self.convert_service(
            service=service,
            images=images,
            container_group_name=container_group_name,
            location=location,
            image_registry_credentials=image_registry_credentials,
            diagnostics=diagnostics,
            os_type=os_type,
            sku=sku,
        )
        existing_service.init_containers = new_service.init_containers
        existing_service.containers = new_service.containers
        existing_service.os_type = new_service.os_type
        existing_service.volumes = new_service.volumes
        existing_service.ip_address = new_service.ip_address
        existing_service.image_registry_credentials = (
            new_service.image_registry_credentials
        )
        existing_service.diagnostics = new_service.diagnostics
        existing_service.restart_policy = new_service.restart_policy
        existing_service.sku = new_service.sku
        return existing_service

    def convert_service(
        self,
        service: ServiceDefinition,
        images: list[str],
        container_group_name: str,
        location: str,
        image_registry_credentials: (
            list[ImageRegistryCredential] | None
        ) = None,
        diagnostics: ContainerGroupDiagnostics | None = None,
        os_type: str | OperatingSystemTypes = OperatingSystemTypes.LINUX,
        sku: str | ContainerGroupSku | None = None,
    ) -> ContainerGroup:
        containers: list[ACIContainer] = []
        init_containers: list[InitContainerDefinition] = []
        for container, image in zip(service.containers, images):
            if container.type == "init":
                init_containers.append(
                    self._convert_init_container(
                        container=container,
                        image=image,
                    )
                )
            else:
                containers.append(
                    self._convert_container(
                        container=container,
                        image=image,
                    )
                )
        return ContainerGroup(
            containers=containers,
            init_containers=init_containers,
            location=location,
            os_type=os_type,
            volumes=self._convert_volumes(service.volumes),
            ip_address=self._convert_ingress(
                ingress=service.ingress,
                dns_name_label=container_group_name,
            ),
            image_registry_credentials=image_registry_credentials,
            diagnostics=diagnostics,
            restart_policy=service.restart_policy,
            sku=sku,
        )

    def _convert_ingress(
        self, ingress: Ingress | None, dns_name_label: str
    ) -> IpAddress:
        if not ingress:
            return IpAddress(
                type=ContainerGroupIpAddressType.PUBLIC,
                ports=[ACIPort(port=80, protocol="TCP")],
                dns_name_label=dns_name_label,
            )
        return IpAddress(
            type=(
                ContainerGroupIpAddressType.PUBLIC
                if ingress.external
                else ContainerGroupIpAddressType.PRIVATE
            ),
            ports=[
                ACIPort(
                    port=int(ingress.port) if ingress.port else 80,
                    protocol=ingress.transport or "TCP",
                )
            ],
            dns_name_label=dns_name_label,
        )

    def _convert_volumes(
        self,
        volumes: list[Volume],
    ) -> list[ACIVolume]:
        result: list[ACIVolume] = []
        for volume in volumes:
            if volume.type == "emptyDir":
                result.append(
                    ACIVolume(
                        name=volume.name,
                        empty_dir={},
                    )
                )
        return result

    def _convert_init_container(
        self,
        container: Container,
        image: str,
    ) -> InitContainerDefinition:
        return InitContainerDefinition(
            name=container.name,
            image=image,
            command=(container.command or []) + (container.args or []),
            environment_variables=self._convert_env(container.env),
            volume_mounts=self._convert_volume_mounts(container.volume_mounts),
        )

    def _convert_container(
        self,
        container: Container,
        image: str,
    ) -> ACIContainer:
        return ACIContainer(
            name=container.name,
            image=image,
            resources=self._convert_resource_requirements(
                container.resources or ResourceRequirements()
            ),
            command=(container.command or []) + (container.args or []),
            environment_variables=self._convert_env(container.env),
            ports=self._convert_ports(container.ports),
            volume_mounts=self._convert_volume_mounts(container.volume_mounts),
            liveness_probe=self._convert_probe(
                container.probes.liveness_probe if container.probes else None
            ),
            readiness_probe=self._convert_probe(
                container.probes.readiness_probe if container.probes else None
            ),
            security_context=self._convert_security_context(
                container.security_context
            ),
        )

    def _convert_security_context(
        self,
        security_context: SecurityContext | None,
    ) -> SecurityContextDefinition | None:
        if not security_context:
            return None
        return SecurityContextDefinition(
            capabilities=SecurityContextCapabilitiesDefinition(
                add=(
                    security_context.capabilities.get("add", [])
                    if security_context.capabilities
                    else []
                ),
                drop=(
                    security_context.capabilities.get("drop", [])
                    if security_context.capabilities
                    else []
                ),
            ),
            run_as_user=security_context.run_as_user,
            run_as_group=security_context.run_as_group,
            privileged=security_context.privileged,
            allow_privilege_escalation=(
                security_context.allow_privilege_escalation
            ),
        )

    def _convert_env(self, env: list[EnvVar]) -> list[EnvironmentVariable]:
        return [
            EnvironmentVariable(name=var.name, value=str(var.value))
            for var in env
        ]

    def _convert_ports(
        self,
        ports: list[Port],
    ) -> list[ContainerPort]:
        return [
            ContainerPort(port=port.container_port, protocol=port.protocol)
            for port in ports
        ]

    def _convert_resource_requirements(
        self,
        resources: ResourceRequirements,
    ) -> ACIResourceRequirements:
        if resources.requests:
            requests = ACIResourceRequests(
                cpu=resources.requests.cpu or 1.0,
                memory_in_gb=(
                    float(resources.requests.memory or 512) / 1024.0
                ),
            )
            if resources.requests.gpu:
                requests.gpu = ACIGpuResource(
                    count=resources.requests.gpu.count,
                    sku=resources.requests.gpu.type or GpuSku.K80,
                )
        else:
            requests = ACIResourceRequests(
                cpu=1.0,
                memory_in_gb=0.5,
                gpu=None,
            )
        if resources.limits:
            limits = ACIResourceLimits(
                cpu=float(resources.limits.cpu or "1"),
                memory_in_gb=(float(resources.limits.memory or 512) / 1024.0),
            )
            if resources.limits.gpu:
                limits.gpu = ACIGpuResource(
                    count=resources.limits.gpu.count,
                    sku=resources.limits.gpu.type or GpuSku.K80,
                )
        else:
            limits = None
        return ACIResourceRequirements(requests=requests, limits=limits)

    def _convert_volume_mounts(
        self,
        volume_mounts: list[VolumeMount],
    ) -> list[ACIVolumeMount] | None:
        return [
            ACIVolumeMount(
                name=mount.name,
                mount_path=mount.mount_path,
                read_only=mount.read_only,
            )
            for mount in volume_mounts
        ]

    def _convert_probe(self, probe: Probe | None) -> ContainerProbe | None:
        if not probe:
            return None
        args: dict = {}
        if probe.http_get:
            args["http_get"] = self._convert_http_get_action(probe.http_get)
        elif probe.exec:
            args["exec_property"] = self._convert_exec_action(probe.exec)
        return ContainerProbe(
            initial_delay_seconds=probe.initial_delay_seconds,
            timeout_seconds=probe.timeout_seconds,
            period_seconds=probe.period_seconds,
            failure_threshold=probe.failure_threshold,
            success_threshold=probe.success_threshold,
            **args,
        )

    def _convert_http_get_action(
        self, action: HTTPGetAction
    ) -> ContainerHttpGet:
        headers = [
            ACIHttpHeader(name=header.name, value=header.value)
            for header in action.http_headers
        ]
        return ContainerHttpGet(
            port=int(action.port),
            path=action.path,
            scheme=action.scheme,
            http_headers=headers,
        )

    def _convert_exec_action(self, action: ExecAction) -> ContainerExec:
        return ContainerExec(
            command=action.command,
        )


class ResultConverter:
    def convert_service_item(
        self,
        service: ContainerGroup,
    ) -> ServiceItem:
        return ServiceItem(
            name=service.name or "unknown",
            uri=self._get_service_uri(service),
            service=self.convert_service(service),
        )

    def convert_service(self, service: ContainerGroup) -> ServiceDefinition:
        result = ServiceDefinition(
            name=service.name,
            ingress=self._convert_ingress(service.ip_address),
            volumes=self._convert_volumes(service.volumes),
            restart_policy=self._convert_restart_policy(
                service.restart_policy
            ),
            latest_ready_revision="latest",
            latest_created_revision="latest",
            traffic=[TrafficAllocation(revision="latest", percent=100)],
        )
        if service.init_containers:
            result.containers = [
                self._convert_init_container(container)
                for container in service.init_containers
            ]
        if service.containers:
            result.containers += [
                self._convert_container(container)
                for container in service.containers
            ]
        return result

    def _convert_restart_policy(
        self,
        restart_policy: str | ContainerGroupRestartPolicy | None,
    ) -> str:
        if isinstance(restart_policy, str):
            return restart_policy
        if isinstance(restart_policy, ContainerGroupRestartPolicy):
            return restart_policy.value
        return "Always"

    def _convert_init_container(
        self,
        container: InitContainerDefinition,
    ) -> Container:
        return Container(
            name=container.name,
            type="init",
            image=container.image,
            command=container.command,
            env=self._convert_env(container.environment_variables),
            volume_mounts=self._convert_volume_mounts(
                container.volume_mounts or []
            ),
        )

    def _convert_container(
        self,
        container: ACIContainer,
    ) -> Container:
        result = Container(
            name=container.name,
            type="main",
            image=container.image,
            command=container.command,
            env=self._convert_env(container.environment_variables),
            ports=self._convert_ports(container.ports),
            resources=self._convert_resource_requirements(container.resources),
            volume_mounts=self._convert_volume_mounts(container.volume_mounts),
            probes=ProbeSet(
                liveness_probe=self._convert_probe(container.liveness_probe),
                readiness_probe=self._convert_probe(container.readiness_probe),
            ),
            security_context=self._convert_security_context(
                container.security_context
            ),
        )
        return result

    def _convert_volume_mounts(
        self,
        volume_mounts: list[ACIVolumeMount] | None,
    ) -> list[VolumeMount]:
        if not volume_mounts:
            return []
        return [
            VolumeMount(
                name=mount.name,
                mount_path=mount.mount_path,
                read_only=mount.read_only,
            )
            for mount in volume_mounts
        ]

    def _convert_ports(
        self,
        ports: list[ContainerPort] | None,
    ) -> list[Port]:
        if not ports:
            return []
        return [
            Port(
                container_port=port.port,
                protocol=self._convert_container_network_protocol(
                    port.protocol
                ),
            )
            for port in ports
        ]

    def _convert_env(
        self,
        env: list[EnvironmentVariable] | None,
    ) -> list[EnvVar]:
        if not env:
            return []
        return [EnvVar(name=var.name, value=var.value) for var in env]

    def _convert_security_context(
        self,
        security_context: SecurityContextDefinition | None,
    ) -> SecurityContext | None:
        if not security_context:
            return None
        result = SecurityContext(
            run_as_user=security_context.run_as_user,
            run_as_group=security_context.run_as_group,
            privileged=security_context.privileged,
            allow_privilege_escalation=(
                security_context.allow_privilege_escalation
            ),
        )
        if security_context.capabilities:
            result.capabilities = {}
            if security_context.capabilities.add:
                result.capabilities["add"] = (
                    security_context.capabilities.add or []
                )
            if security_context.capabilities.drop:
                result.capabilities["drop"] = (
                    security_context.capabilities.drop or []
                )
        return result

    def _convert_container_network_protocol(
        self,
        protocol: str | ContainerNetworkProtocol | None,
    ) -> str:
        if not protocol:
            return "TCP"
        if isinstance(protocol, str):
            return protocol
        if protocol == ContainerNetworkProtocol.TCP:
            return "TCP"
        elif protocol == ContainerNetworkProtocol.UDP:
            return "UDP"
        else:
            raise BadRequestError(f"Unsupported network protocol: {protocol}")

    def _convert_resource_requirements(
        self,
        resources: ACIResourceRequirements | None,
    ) -> ResourceRequirements | None:
        if not resources:
            return None
        requests = ResourceRequests(
            cpu=resources.requests.cpu,
            memory=int(resources.requests.memory_in_gb * 1024),
        )
        if resources.requests.gpu:
            requests.gpu = GPUResource(
                count=resources.requests.gpu.count,
                type=(
                    resources.requests.gpu.sku.value
                    if isinstance(resources.requests.gpu.sku, GpuSku)
                    else resources.requests.gpu.sku
                ),
            )
        if resources.limits:
            limits = ResourceLimits()
            if resources.limits.cpu:
                limits.cpu = resources.limits.cpu
            if resources.limits.memory_in_gb:
                limits.memory = int(resources.limits.memory_in_gb * 1024)
            if resources.limits.gpu:
                limits.gpu = GPUResource(
                    count=resources.limits.gpu.count,
                    type=(
                        resources.limits.gpu.sku.value
                        if isinstance(resources.limits.gpu.sku, GpuSku)
                        else resources.limits.gpu.sku
                    ),
                )
        else:
            limits = None
        return ResourceRequirements(
            requests=requests,
            limits=limits,
        )

    def _convert_ingress(
        self,
        ingress: IpAddress | None,
    ) -> Ingress | None:
        if not ingress:
            return None
        return Ingress(
            external=ingress.type == ContainerGroupIpAddressType.PUBLIC,
            target_port=ingress.ports[0].port if ingress.ports else 80,
            transport=ingress.ports[0].protocol if ingress.ports else "TCP",
        )

    def _convert_volumes(
        self,
        volumes: list[ACIVolume] | None,
    ) -> list[Volume]:
        if not volumes:
            return []
        result: list[Volume] = []
        for volume in volumes:
            if volume.empty_dir:
                result.append(
                    Volume(
                        name=volume.name,
                        type="emptyDir",
                    )
                )
        return result

    def _convert_probe(self, probe: ContainerProbe | None) -> Probe | None:
        if not probe:
            return None
        http_get = None
        exec_action = None
        if probe.http_get:
            http_get = self._convert_http_get_action(probe.http_get)
        elif probe.exec_property:
            exec_action = self._convert_exec_action(probe.exec_property)
        return Probe(
            initial_delay_seconds=probe.initial_delay_seconds,
            timeout_seconds=probe.timeout_seconds,
            period_seconds=probe.period_seconds,
            failure_threshold=probe.failure_threshold,
            success_threshold=probe.success_threshold,
            http_get=http_get,
            exec=exec_action,
        )

    def _convert_http_get_action(
        self, action: ContainerHttpGet
    ) -> HTTPGetAction:
        headers = []
        if action.http_headers:
            for header in action.http_headers:
                if header.name and header.value:
                    headers.append(
                        HTTPHeader(name=header.name, value=header.value)
                    )
        return HTTPGetAction(
            port=str(action.port),
            path=action.path or "/",
            scheme=action.scheme or "HTTP",
            http_headers=headers,
        )

    def _convert_exec_action(self, action: ContainerExec) -> ExecAction:
        return ExecAction(
            command=action.command or [],
        )

    def convert_services(
        self,
        services: list[ContainerGroup],
    ) -> list[ServiceItem]:
        result: list[ServiceItem] = []
        for service in services:
            result.append(self.convert_service_item(service))
        return result

    def _get_service_uri(self, container_group: ContainerGroup) -> str | None:
        """Extract the public URL from the container group response."""
        if container_group.ip_address and container_group.ip_address.fqdn:
            port = ""
            if (
                container_group.ip_address.ports
                and container_group.ip_address.ports[0].port != 80
            ):
                port = f":{container_group.ip_address.ports[0].port}"
            return f"http://{container_group.ip_address.fqdn}{port}"
        elif container_group.ip_address and container_group.ip_address.ip:
            port = ""
            if (
                container_group.ip_address.ports
                and container_group.ip_address.ports[0].port != 80
            ):
                port = f":{container_group.ip_address.ports[0].port}"
            return f"http://{container_group.ip_address.ip}{port}"
        return None



"""
Object Store on File System.
"""

from __future__ import annotations

__all__ = ["FileSystem"]

import os
import shutil
import uuid
from datetime import datetime, timezone
from typing import IO, Any
from urllib.parse import urljoin
from urllib.request import pathname2url

from x8.core import Context, DataModel, Operation, Response
from x8.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    NotModified,
    PreconditionFailedError,
)
from x8.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    OrderBy,
    OrderByDirection,
    OrderByTerm,
    QueryFunction,
)
from x8.storage._common import (
    CollectionResult,
    CollectionStatus,
    MatchCondition,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)
from x8.storage.document_store import DocumentStore, DocumentTransaction
from x8.storage.document_store.providers.sqlite import SQLite

from .._helper import (
    QueryArgs,
    get_collection_config,
    get_query_args,
    get_transfer_config,
)
from .._models import (
    ObjectCollectionConfig,
    ObjectItem,
    ObjectKey,
    ObjectList,
    ObjectProperties,
    ObjectTransferConfig,
    ObjectVersion,
)

DB_FILE_NAME = "__db"
CONFIG_ID = "config"
CONFIG_COLLECTION = "config"
OBJECT_COLLECTION = "object"
OBJECT_DOCUMENT_TYPE = "object"
VERSION_DOCUMENT_TYPE = "version"
PK = "#"


class ObjectDocument(DataModel):
    id: str
    pk: str
    object_id: str
    version: str | None
    metadata: dict | None
    properties: dict
    ts: float
    type: str


class ConfigDocument(DataModel):
    id: str
    pk: str
    versioned: bool


class FileSystem(StoreProvider):
    store_path: str
    folder: str | None
    db_file_name: str
    nparams: dict[str, Any]

    _init: bool
    _db_cache: dict[str, DocumentStore]

    def __init__(
        self,
        store_path: str = ".",
        folder: str | None = None,
        db_file_name: str = DB_FILE_NAME,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            store_path:
                Base store path. Defaults to ".".
            folder:
                Folder name mapped to object store collection name.
            db_file_name:
                File name of the SQLite DB that
                manages the metadata in each collection folder.
            nparams:
                Native parameters to file system operations.
        """
        self.store_path = store_path
        self.folder = folder
        self.nparams = nparams
        self.db_file_name = db_file_name
        self._init = False
        self._db_cache = dict()

    def _get_db(self, collection: str):
        if collection in self._db_cache:
            return self._db_cache[collection]
        collection_path = os.path.join(self.store_path, collection)
        db_path = os.path.join(collection_path, self.db_file_name)
        db = DocumentStore(
            collection=OBJECT_COLLECTION,
            __provider__=SQLite(
                database=db_path,
            ),
        )
        self._db_cache[collection] = db
        return db

    def _is_versioned(self, collection: str) -> bool:
        db = self._get_db(collection)
        try:
            res = db.get(key=CONFIG_ID, collection=CONFIG_COLLECTION)
            return res.result.value["versioned"]
        except NotFoundError:
            return False

    def __setup__(self, context: Context | None = None) -> None:
        if not self._init:
            return

        self._create_folder(self.store_path)
        self._init = True

    def _create_folder(self, folder: str) -> None:
        if not os.path.exists(folder):
            os.makedirs(folder, exist_ok=True)

    def _get_folder_name(self, op_parser: StoreOperationParser) -> str:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        folder = (
            collection_name or self.folder or self.__component__.collection
        )
        if folder is None:
            raise BadRequestError("Collection name should be specified")
        return folder

    def _validate(self, op_parser: StoreOperationParser):
        pass

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = self.create_collection(
                collection=self._get_folder_name(op_parser),
                config=get_collection_config(op_parser),
                exists=op_parser.get_where_exists(),
            )
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = self.drop_collection(
                collection=self._get_folder_name(op_parser),
                exists=op_parser.get_where_exists(),
            )
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = self.list_collections()
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            result = self.has_collection(
                collection=self._get_folder_name(op_parser),
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = self.put(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_value_as_bytes(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
                returning=op_parser.get_returning(),
                config=get_transfer_config(op_parser),
                collection=self._get_folder_name(op_parser),
            )
        # GET value
        elif op_parser.op_equals(StoreOperation.GET):
            result = self.get(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                match_condition=op_parser.get_match_condition(),
                start=op_parser.get_start(),
                end=op_parser.get_end(),
                config=get_transfer_config(op_parser),
                collection=self._get_folder_name(op_parser),
            )
        # GET metadata or properties
        elif op_parser.op_equals(
            StoreOperation.GET_METADATA
        ) or op_parser.op_equals(StoreOperation.GET_PROPERTIES):
            result = self.get_properties(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=op_parser.get_match_condition(),
                collection=self._get_folder_name(op_parser),
            )
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET_VERSIONS):
            result = self.get_versions(
                id=op_parser.get_id_as_str(),
                collection=self._get_folder_name(op_parser),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = self.delete(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=op_parser.get_match_condition(),
                collection=self._get_folder_name(op_parser),
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            result = self.update(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
                collection=self._get_folder_name(op_parser),
            )
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            result = self.copy(
                id=op_parser.get_id_as_str(),
                source_id=op_parser.get_source_id_as_str(),
                source_version=op_parser.get_source_version(),
                source_collection=op_parser.get_source_collection(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
                collection=self._get_folder_name(op_parser),
            )
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            result = self.generate(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                method=op_parser.get_method(),
                expiry=op_parser.get_expiry_in_seconds(),
                collection=self._get_folder_name(op_parser),
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            query_args = get_query_args(op_parser)
            result = self.query(
                query_args=query_args,
                collection=self._get_folder_name(op_parser),
            )
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            query_args = get_query_args(op_parser)
            result = self.count(
                query_args=query_args,
                collection=self._get_folder_name(op_parser),
            )
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = self.batch(
                op_parser.get_operation_parsers(),
                collection=self._get_folder_name(op_parser),
            )
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            result = self.close()
        return Response(result=result)

    def create_collection(
        self,
        collection: str,
        config: ObjectCollectionConfig | None,
        exists: bool | None,
    ) -> CollectionResult:
        db = self._get_db(collection)
        folder_path = os.path.join(self.store_path, collection)
        if os.path.isdir(folder_path):
            if exists is False:
                raise ConflictError
            else:
                return CollectionResult(status=CollectionStatus.EXISTS)
        self._create_folder(folder_path)
        db.create_collection()
        db.create_collection(CONFIG_COLLECTION)
        versioned = False
        if config is not None and config.versioned:
            versioned = True
        db.put(
            ConfigDocument(id=CONFIG_ID, pk=CONFIG_ID, versioned=versioned),
            collection=CONFIG_COLLECTION,
        )
        return CollectionResult(status=CollectionStatus.CREATED)

    def drop_collection(
        self,
        collection: str,
        exists: bool | None,
    ) -> CollectionResult:
        folder_path = os.path.join(self.store_path, collection)
        if not os.path.isdir(folder_path):
            if exists is True:
                raise NotFoundError
            else:
                return CollectionResult(status=CollectionStatus.NOT_EXISTS)
        db = self._get_db(collection)
        db.close()
        if collection in self._db_cache:
            self._db_cache.pop(collection)
        import shutil

        shutil.rmtree(folder_path)
        return CollectionResult(status=CollectionStatus.DROPPED)

    def list_collections(self) -> list[str]:
        result = []
        items = os.listdir(self.store_path)
        for item in items:
            full_path = os.path.join(self.store_path, item)
            if os.path.isdir(full_path):
                result.append(item)
        return result

    def has_collection(self, collection: str) -> bool:
        folder_path = os.path.join(self.store_path, collection)
        return os.path.isdir(folder_path)

    def put(
        self,
        id: str,
        value: bytes | None,
        file: str | None,
        stream: IO | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        returning: str | None,
        config: ObjectTransferConfig | None,
        collection: str,
    ) -> ObjectItem:
        db = self._get_db(collection)
        versioned = self._is_versioned(collection)
        version = None
        link_path = None
        try:
            current_item = self._db_get(db, id, version)
        except NotFoundError:
            current_item = None
        self._match(current_item, match_condition)

        if versioned:
            version = str(uuid.uuid4())
        object_path, link_path = self._convert_object_link_path(
            collection, id, version
        )
        os.makedirs(os.path.dirname(object_path), exist_ok=True)
        if file:
            shutil.copy(file, object_path)
        elif value:
            with open(object_path, "wb") as f:
                f.write(value)
        elif stream:
            with open(object_path, "wb") as f:
                shutil.copyfileobj(stream, f)

        if link_path:
            os.makedirs(os.path.dirname(link_path), exist_ok=True)
            if os.path.islink(link_path):
                os.unlink(link_path)
            os.symlink(object_path, link_path)

        last_modified = datetime.now(timezone.utc).timestamp()
        content_length = os.path.getsize(object_path)
        ts = datetime.now(timezone.utc).timestamp()
        if properties is None:
            props = dict()
        else:
            props = properties.copy()
        props["last_modified"] = last_modified
        props["content_length"] = content_length
        if version:
            transaction = DocumentTransaction()
            transaction.put(
                value=ObjectDocument(
                    id=self._get_db_object_id(id),
                    pk=PK,
                    object_id=id,
                    version=version,
                    metadata=metadata,
                    properties=props,
                    ts=ts,
                    type=OBJECT_DOCUMENT_TYPE,
                )
            )
            transaction.put(
                value=ObjectDocument(
                    id=self._get_db_version_id(id, version),
                    pk=PK,
                    object_id=id,
                    version=version,
                    metadata=metadata,
                    properties=props,
                    ts=ts,
                    type=VERSION_DOCUMENT_TYPE,
                )
            )
            res = db.transact(transaction=transaction)
            etag = res.result[1].properties.etag
        else:
            res = db.put(
                ObjectDocument(
                    id=self._get_db_object_id(id),
                    pk=PK,
                    object_id=id,
                    version=version,
                    metadata=metadata,
                    properties=props,
                    ts=ts,
                    type=OBJECT_DOCUMENT_TYPE,
                )
            )
            etag = res.result.properties.etag
        return_value = None
        if returning == "new":
            return_value = value
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            value=return_value,
            metadata=metadata,
            properties=ObjectProperties(
                etag=etag,
                last_modified=last_modified,
                content_length=content_length,
            ),
            url=self._convert_url(link_path or object_path),
        )

    def get(
        self,
        id: str,
        version: str | None,
        file: str | None,
        stream: IO | None,
        match_condition: MatchCondition,
        start: int | None,
        end: int | None,
        config: ObjectTransferConfig | None,
        collection: str,
    ) -> ObjectItem:
        db = self._get_db(collection)
        value = None
        object_path, link_path = self._convert_object_link_path(
            collection, id, version
        )

        if not os.path.isfile(object_path) and not os.path.islink(object_path):
            raise NotFoundError

        item = self._db_get(db, id, version)
        self._match(item, match_condition)

        with open(object_path, "rb") as f:
            if start is not None or end is not None:
                f.seek(start or 0)
                data = f.read(
                    (end - (start or 0) + 1) if end is not None else None
                )
            else:
                data = f.read()
        if file:
            os.makedirs(os.path.dirname(file), exist_ok=True)
            with open(file, "wb") as out_file:
                out_file.write(data)
        elif stream:
            stream.write(data)
            stream.seek(0)
        else:
            value = data

        return ObjectItem(
            key=ObjectKey(id=id, version=item.version),
            value=value,
            metadata=item.metadata,
            properties=ObjectProperties.from_dict(item.properties),
            url=self._convert_url(link_path or object_path),
        )

    def get_properties(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
        collection: str,
    ) -> ObjectItem:
        db = self._get_db(collection)
        object_path, link_path = self._convert_object_link_path(
            collection, id, version
        )
        try:
            item = self._db_get(db, id, version)
            self._match(item, match_condition)
        except NotFoundError:
            raise NotFoundError
        return ObjectItem(
            key=ObjectKey(id=id, version=item.version),
            metadata=item.metadata,
            properties=ObjectProperties.from_dict(item.properties),
            url=self._convert_url(link_path or object_path),
        )

    def get_versions(self, id: str, collection: str) -> ObjectItem:
        db = self._get_db(collection)
        object_path, link_path = self._convert_object_link_path(
            collection, id, None
        )
        res = db.query(
            where=And(
                lexpr=Comparison(
                    lexpr=Field(path="object_id"),
                    op=ComparisonOp.EQ,
                    rexpr=id,
                ),
                rexpr=Comparison(
                    lexpr=Field(path="type"),
                    op=ComparisonOp.EQ,
                    rexpr=VERSION_DOCUMENT_TYPE,
                ),
            ),
            order_by=OrderBy(
                terms=[
                    OrderByTerm(
                        field="ts",
                        direction=OrderByDirection.DESC,
                    )
                ]
            ),
        )

        versions: list[ObjectVersion] = []
        for item in res.result.items:
            obj = ObjectDocument.from_dict(item.value)
            properties = ObjectProperties.from_dict(obj.properties)
            properties.etag = item.properties.etag
            versions.append(
                ObjectVersion(
                    version=obj.version,
                    properties=properties,
                    latest=False,
                )
            )
        if len(versions) == 0:
            raise NotFoundError
        versions[0].latest = True
        return ObjectItem(
            key=ObjectKey(id=id),
            versions=versions[::-1],
            url=self._convert_url(link_path or object_path),
        )

    def delete(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
        collection: str,
    ):
        db = self._get_db(collection)
        versioned = self._is_versioned(collection)
        try:
            current_item = self._db_get(db, id, version)
        except NotFoundError:
            current_item = None
        self._match(current_item, match_condition)
        if versioned:
            try:
                obj_res = db.get(key=self._get_db_object_key(id))
            except NotFoundError:
                raise NotFoundError
            if (
                version == "*"
                or version is None
                or version == obj_res.result.value["version"]
            ):
                res = db.query(
                    where=And(
                        lexpr=Comparison(
                            lexpr=Field(path="object_id"),
                            op=ComparisonOp.EQ,
                            rexpr=id,
                        ),
                        rexpr=Comparison(
                            lexpr=Field(path="type"),
                            op=ComparisonOp.EQ,
                            rexpr=VERSION_DOCUMENT_TYPE,
                        ),
                    )
                )
                for item in res.result.items:
                    obj = ObjectDocument.from_dict(item.value)
                    object_path, link_path = self._convert_object_link_path(
                        collection, obj.object_id, obj.version
                    )
                    if os.path.isfile(object_path) or os.path.islink(
                        object_path
                    ):
                        self._delete_file(object_path)
                    if link_path and os.path.islink(link_path):
                        self._delete_file(link_path)
                    db.delete(key=item.key)
                db.delete(key=obj_res.result.key)
            else:
                res = db.get(key=self._get_db_version_key(id, version))
                obj = ObjectDocument.from_dict(res.result.value)
                object_path, link_path = self._convert_object_link_path(
                    collection, obj.object_id, obj.version
                )
                self._delete_file(object_path)
                if link_path and os.path.islink(link_path):
                    self._delete_file(link_path)
                db.delete(key=res.result.key)
        else:
            try:
                res = db.get(key=self._get_db_object_key(id))
            except NotFoundError:
                raise NotFoundError
            obj = ObjectDocument.from_dict(res.result.value)
            object_path, link_path = self._convert_object_link_path(
                collection, obj.object_id, obj.version
            )
            self._delete_file(object_path)
            db.delete(key=res.result.key)
        return None

    def update(
        self,
        id: str,
        version: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        collection: str,
    ) -> ObjectItem:
        db = self._get_db(collection)
        object_path, link_path = self._convert_object_link_path(
            collection, id, version
        )
        versioned = self._is_versioned(collection)
        last_modified = datetime.now(timezone.utc).timestamp()
        try:
            obj_res = db.get(key=self._get_db_object_key(id))
        except NotFoundError:
            raise NotFoundError
        obj = ObjectDocument.from_dict(obj_res.result.value)
        if versioned:
            if obj.version and (version is None or version == obj.version):
                ver_res = db.get(key=self._get_db_version_key(id, obj.version))
                ver = ObjectDocument.from_dict(ver_res.result.value)
                ver.properties["etag"] = ver_res.result.properties.etag
                self._match(ver, match_condition)
                ver.properties.pop("etag")
                obj.properties["last_modified"] = last_modified
                obj.properties = obj.properties | (properties or {})
                ver.properties["last_modified"] = last_modified
                ver.properties = ver.properties | (properties or {})
                if metadata:
                    obj.metadata = metadata
                    ver.metadata = metadata
                transaction = DocumentTransaction()
                transaction.put(obj)
                transaction.put(ver)
                res = db.transact(transaction=transaction)
                etag = res.result[1].properties.etag
                item = obj
            elif version is not None:
                ver_res = db.get(key=self._get_db_version_key(id, version))
                ver = ObjectDocument.from_dict(ver_res.result.value)
                ver.properties["etag"] = ver_res.result.properties.etag
                self._match(ver, match_condition)
                ver.properties.pop("etag")
                ver.properties["last_modified"] = last_modified
                ver.properties = ver.properties | (properties or {})
                if metadata:
                    ver.metadata = metadata
                res = db.put(ver)
                etag = res.result.properties.etag
                item = ver
        else:
            obj.properties["etag"] = obj_res.result.properties.etag
            self._match(obj, match_condition)
            obj.properties.pop("etag")
            obj.properties["last_modified"] = last_modified
            obj.properties = obj.properties | (properties or {})
            if metadata:
                obj.metadata = metadata
            res = db.put(obj)
            etag = res.result.properties.etag
            item = obj
        item.properties["etag"] = etag
        return ObjectItem(
            key=ObjectKey(id=id, version=item.version),
            metadata=item.metadata,
            properties=ObjectProperties.from_dict(item.properties),
            url=self._convert_url(link_path or object_path),
        )

    def copy(
        self,
        id: str,
        source_id: str,
        source_version: str | None,
        source_collection: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        collection: str,
    ) -> ObjectItem:
        source_collection = source_collection or collection
        source_db = self._get_db(source_collection)
        dest_db = self._get_db(collection)

        source_object_path, source_link_path = self._convert_object_link_path(
            source_collection,
            source_id,
            source_version,
        )

        if not os.path.isfile(source_object_path) and not os.path.islink(
            source_object_path
        ):
            raise NotFoundError
        source_item = self._db_get(
            source_db,
            source_id,
            source_version,
        )
        dest_item = None
        try:
            dest_item = self._db_get(dest_db, id, None)
        except NotFoundError:
            pass
        self._match(dest_item, match_condition)
        new_properties = source_item.properties | (properties or {})
        new_properties.pop("etag")
        new_properties.pop("last_modified")
        if source_item.metadata or metadata:
            new_metadata = (source_item.metadata or {}) | (metadata or {})
        else:
            new_metadata = None
        return self.put(
            id=id,
            value=None,
            file=source_object_path,
            stream=None,
            metadata=new_metadata,
            properties=new_properties,
            match_condition=match_condition,
            returning=None,
            config=None,
            collection=collection,
        )

    def generate(
        self,
        id: str,
        version: str | None,
        method: str | None,
        expiry: int | None,
        collection: str,
    ) -> ObjectItem:
        object_path, link_path = self._convert_object_link_path(
            collection, id, version
        )
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            url=self._convert_url(link_path or object_path),
        )

    def query(
        self,
        query_args: QueryArgs,
        collection: str,
    ) -> ObjectList:
        db = self._get_db(collection)
        where: Expression = Comparison(
            lexpr=Field(path="type"),
            op=ComparisonOp.EQ,
            rexpr=OBJECT_DOCUMENT_TYPE,
        )
        if query_args.prefix:
            where = And(
                lexpr=where,
                rexpr=QueryFunction.starts_with(
                    field="object_id",
                    value=query_args.prefix,
                ),
            )
        if query_args.continuation:
            where = And(
                lexpr=where,
                rexpr=Comparison(
                    lexpr=Field(path="object_id"),
                    op=ComparisonOp.GT,
                    rexpr=query_args.continuation,
                ),
            )
        limit = query_args.limit
        if query_args.paging and query_args.page_size:
            if limit:
                if query_args.page_size < limit:
                    limit = query_args.page_size
            else:
                limit = query_args.page_size

        items: list[ObjectItem] = []
        prefixes: list[str] = []
        continuation = None
        res = db.query(
            where=where,
            order_by=OrderBy(
                terms=[
                    OrderByTerm(
                        field="object_id",
                        direction=OrderByDirection.ASC,
                    )
                ]
            ),
            limit=limit,
        )
        for item in res.result.items:
            obj = ObjectDocument.from_dict(item.value)
            continuation = obj.object_id
            if query_args.delimiter:
                relative_path = obj.object_id
                if query_args.prefix:
                    start = len(query_args.prefix)
                    relative_path = relative_path[start:]
                prefix_part, _, rest = relative_path.partition(
                    query_args.delimiter
                )
                if rest:
                    prefix = (
                        (query_args.prefix or "")
                        + prefix_part
                        + query_args.delimiter
                    )
                    if prefix not in prefixes:
                        prefixes.append(prefix)
                    continue
            object_path, link_path = self._convert_object_link_path(
                collection, obj.object_id, obj.version
            )
            items.append(
                ObjectItem(
                    key=ObjectKey(id=obj.object_id, version=obj.version),
                    url=self._convert_url(link_path or object_path),
                )
            )
        if not query_args.paging:
            continuation = None
        if limit:
            if len(res.result.items) < limit:
                continuation = None
        return ObjectList(
            items=items,
            prefixes=sorted(prefixes),
            continuation=continuation,
        )

    def count(
        self,
        query_args: QueryArgs,
        collection: str,
    ) -> int:
        res = self.query(query_args, collection)
        count = len(res.items)
        if res.prefixes:
            count = count + len(res.prefixes)
        return count

    def batch(
        self,
        op_parsers: list[StoreOperationParser],
        collection: str,
    ) -> list[Any]:
        result: list = []
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.DELETE):
                id = op_parser.get_id_as_str()
                version = op_parser.get_version()
                match_condition = op_parser.get_match_condition()
                self.delete(id, version, match_condition, collection)
                result.append(None)
        return result

    def close(self) -> Any:
        pass

    def _convert_url(self, file_path: str) -> str:
        absolute_path = os.path.abspath(file_path)
        url_path = pathname2url(absolute_path)
        return urljoin("file://", url_path)

    def _convert_object_link_path(
        self, collection: str, id: str, version: str | None
    ) -> tuple[str, str | None]:
        link_path = None
        if version:
            link_path = os.path.join(
                self.store_path,
                collection,
                id.lstrip("/"),
            )
            object_path = os.path.join(
                self.store_path,
                collection,
                version,
                id.lstrip("/"),
            )
        else:
            object_path = os.path.join(
                self.store_path,
                collection,
                id.lstrip("/"),
            )
        object_path = os.path.normpath(object_path)
        if link_path is not None:
            link_path = os.path.normpath(link_path)
        return object_path, link_path

    def _db_get(
        self,
        db: DocumentStore,
        id: str,
        version: str | None,
    ) -> ObjectDocument:
        try:
            get_res = db.get(
                key=(
                    self._get_db_object_key(id)
                    if version is None
                    else self._get_db_version_key(id, version)
                )
            )
            item = ObjectDocument.from_dict(get_res.result.value)
            if item.version:
                get_res = db.get(
                    key=(self._get_db_version_key(id, item.version))
                )
                item = ObjectDocument.from_dict(get_res.result.value)
            etag = (
                get_res.result.properties.etag
                if get_res.result.properties is not None
                else None
            )
            item.properties["etag"] = etag
            return item
        except NotFoundError:
            raise NotFoundError

    def _match(
        self, item: ObjectDocument | None, match_condition: MatchCondition
    ):
        if item is None:
            if (
                match_condition.if_match
                or match_condition.if_unmodified_since
                or match_condition.exists is True
            ):
                raise PreconditionFailedError
            return

        if match_condition.exists is False:
            raise PreconditionFailedError

        etag = item.properties["etag"]
        last_modified = item.properties["last_modified"]
        version = item.version
        if match_condition.if_match is not None:
            if (
                match_condition.if_match != etag
                and match_condition.if_match != "*"
            ):
                raise PreconditionFailedError
        if match_condition.if_none_match is not None:
            etag_list = [
                etag.strip()
                for etag in match_condition.if_none_match.split(",")
            ]
            if etag in etag_list:
                raise NotModified
        if match_condition.if_version_match is not None:
            if match_condition.if_version_match != version:
                raise PreconditionFailedError
        if match_condition.if_version_not_match is not None:
            if match_condition.if_version_not_match == version:
                raise PreconditionFailedError
        if match_condition.if_modified_since is not None:
            if last_modified <= match_condition.if_modified_since:
                raise NotModified
        if match_condition.if_unmodified_since is not None:
            if last_modified > match_condition.if_unmodified_since:
                raise PreconditionFailedError

    def _delete_file(self, path):
        os.remove(path)
        folder = os.path.dirname(path)
        store_abs_path = os.path.abspath(self.store_path)
        while folder and os.path.abspath(folder) != store_abs_path:
            if os.path.isdir(folder):
                try:
                    os.rmdir(folder)
                    folder = os.path.dirname(folder)
                except OSError:
                    break
            else:
                break

    def _get_db_object_key(self, id: str) -> dict:
        return {"id": self._get_db_object_id(id), "pk": PK}

    def _get_db_version_key(self, id: str, version: str) -> dict:
        return {"id": self._get_db_version_id(id, version), "pk": PK}

    def _get_db_object_id(self, id: str) -> str:
        return f"{OBJECT_DOCUMENT_TYPE}-{id}"

    def _get_db_version_id(self, id: str, version: str) -> str:
        return f"{VERSION_DOCUMENT_TYPE}-{id}-{version}"



"""
AWS App Runner deployment.
"""

from __future__ import annotations

__all__ = ["AWSAppRunner"]

import json
import time
from typing import Any

import boto3
from botocore.exceptions import ClientError
from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Context, OperationParser, Response, RunContext
from x8.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import Expression

from .._feature import ContainerDeploymentFeature
from .._helper import map_images, requires_container_registry
from .._models import (
    Container,
    EnvVar,
    HTTPGetAction,
    Ingress,
    Port,
    Probe,
    ProbeSet,
    ResourceRequests,
    ResourceRequirements,
    Revision,
    Scale,
    ServiceDefinition,
    ServiceItem,
    ServiceOverlay,
    TCPSocketAction,
    TrafficAllocation,
)
from ._base import BaseContainerDeploymentProvider


class AWSAppRunner(BaseContainerDeploymentProvider):
    region: str
    service_name: str | None

    ecr_connection_arn: str | None = None
    auto_scaling_configuration_arn: str | None = None
    vpc_connector_arn: str | None = None
    authentication_configuration_access_role_arn: str | None = None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    profile_name: str | None
    nparams: dict[str, Any]

    _client: Any
    _iam_client: Any
    _init: bool = False
    _op_converter: OperationConverter
    _result_converter: ResultConverter

    def __init__(
        self,
        region: str = "us-west-2",
        service_name: str | None = None,
        service: ServiceDefinition | None = None,
        service_override: ServiceOverlay | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        ecr_connection_arn: str | None = None,
        auto_scaling_configuration_arn: str | None = None,
        vpc_connector_arn: str | None = None,
        authentication_configuration_access_role_arn: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        profile_name: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize AWS App Runner deployment provider.

        Args:
            region: AWS region for the App Runner service.
            service_name: The name of the App Runner service.
            service: Service definition to deploy.
            service_override: Service override definition to apply.
            containerizer: Containerizer instance for building images.
            container_registry: Container registry instance for pushing images.
            ecr_connection_arn: ARN of the ECR connection (if any).
            auto_scaling_configuration_arn:
                ARN of the auto scaling configuration (if any).
            vpc_connector_arn: ARN of the VPC connector to use (if any).
            authentication_configuration_access_role_arn:
                ARN of the IAM role for App Runner authentication.
            aws_access_key_id: AWS access key ID.
            aws_secret_access_key: AWS secret access key.
            aws_session_token: AWS session token.
            profile_name: AWS profile name to use.
            nparams: Additional parameters for the AWS client.
        """
        self.region = region
        self.service_name = service_name
        self.service = service
        self.overlay = service_override
        self.containerizer = containerizer
        self.container_registry = container_registry
        self.ecr_connection_arn = ecr_connection_arn
        self.auto_scaling_configuration_arn = auto_scaling_configuration_arn
        self.vpc_connector_arn = vpc_connector_arn
        self.authentication_configuration_access_role_arn = (
            authentication_configuration_access_role_arn
        )
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.profile_name = profile_name
        self.nparams = nparams

        self._init = False
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()

        super().__init__(**kwargs)

    def __supports__(self, feature):
        return feature not in [
            ContainerDeploymentFeature.MULTIPLE_REVISIONS,
            ContainerDeploymentFeature.REVISION_DELETE,
            ContainerDeploymentFeature.MULTIPLE_CONTAINERS,
        ]

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        session_kwargs = {}
        if self.aws_access_key_id:
            session_kwargs["aws_access_key_id"] = self.aws_access_key_id
        if self.aws_secret_access_key:
            session_kwargs["aws_secret_access_key"] = (
                self.aws_secret_access_key
            )
        if self.aws_session_token:
            session_kwargs["aws_session_token"] = self.aws_session_token
        if self.profile_name:
            session_kwargs["profile_name"] = self.profile_name

        session = boto3.Session(**session_kwargs)
        self._client = session.client(
            "apprunner", region_name=self.region, **self.nparams
        )
        self._iam_client = session.client(
            "iam", region_name=self.region, **self.nparams
        )
        self._init = True

    def _ensure_container_registry(
        self, service: ServiceDefinition
    ) -> ContainerRegistry | None:
        container_registry = self._get_container_registry()
        if container_registry:
            return container_registry
        if not requires_container_registry(service):
            return None

        from x8.compute.container_registry import ContainerRegistry
        from x8.compute.container_registry.providers.amazon_elastic_container_registry import (  # noqa
            AmazonElasticContainerRegistry,
        )

        return ContainerRegistry(
            __provider__=AmazonElasticContainerRegistry(
                region=self.region,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                profile_name=self.profile_name,
            ),
        )

    def create_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        self.__setup__()
        where_exists = OperationParser.parse_where_exists(where)
        service_def = self._normalize_service_definition(service)
        app_runner_service_name = self._get_service_name(service_def)
        ecr_connection_arn = self.ecr_connection_arn
        auto_scaling_configuration_arn = self.auto_scaling_configuration_arn

        app_runner_service = self._get_service_by_name(app_runner_service_name)

        if app_runner_service:
            if where_exists is False:
                raise PreconditionFailedError(
                    f"Service {app_runner_service_name} already exists."
                )

            # Update existing service
            images = map_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service=service_def
                ),
                run_context=run_context,
            )

            if not auto_scaling_configuration_arn:
                auto_scaling_configuration_arn = (
                    self._get_or_create_auto_scaling_configuration(
                        service_name=app_runner_service_name,
                        scale=service_def.scale,
                    )
                )

            update_params = self._op_converter.convert_service_update(
                service_arn=app_runner_service["ServiceArn"],
                service=service_def,
                images=images,
                auto_scaling_config_arn=auto_scaling_configuration_arn,
                ecr_connection_arn=ecr_connection_arn,
                vpc_connector_arn=self.vpc_connector_arn,
            )

            response = self._client.update_service(**update_params)
            operation_id = response["OperationId"]

            # Wait for operation to complete
            self._wait_for_creation(
                operation_id, response["Service"]["ServiceArn"]
            )

            # Get updated service
            updated_service = self._client.describe_service(
                ServiceArn=response["Service"]["ServiceArn"]
            )
            result = self._result_converter.convert_service_item(
                updated_service["Service"]
            )
            print(f"Service updated successfully: {result.uri}")
        else:
            if where_exists is True:
                raise PreconditionFailedError(
                    f"Service {app_runner_service_name} not found."
                )

            # Create new service
            images = map_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service=service_def
                ),
                run_context=run_context,
            )

            if not auto_scaling_configuration_arn:
                auto_scaling_configuration_arn = (
                    self._get_or_create_auto_scaling_configuration(
                        service_name=app_runner_service_name,
                        scale=service_def.scale,
                    )
                )

            authentication_configuration_access_role_arn = (
                self.authentication_configuration_access_role_arn
            )
            if not authentication_configuration_access_role_arn:
                authentication_configuration_access_role_arn = (
                    self._create_ecr_access_role()
                )
            create_params = self._op_converter.convert_service_create(
                service_name=app_runner_service_name,
                service=service_def,
                images=images,
                auto_scaling_config_arn=auto_scaling_configuration_arn,
                ecr_connection_arn=ecr_connection_arn,
                vpc_connector_arn=self.vpc_connector_arn,
                authentication_configuration_access_role_arn=authentication_configuration_access_role_arn,  # noqa
            )

            response = self._client.create_service(**create_params)
            operation_id = response["OperationId"]

            # Wait for operation to complete
            self._wait_for_creation(
                operation_id, response["Service"]["ServiceArn"]
            )

            # Get created service
            created_service = self._client.describe_service(
                ServiceArn=response["Service"]["ServiceArn"]
            )
            result = self._result_converter.convert_service_item(
                created_service["Service"]
            )
            print(f"Service created successfully: {result.uri}")

        return Response(result=result)

    async def acreate_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        return self.create_service(
            service=service,
            where=where,
            run_context=run_context,
            **kwargs,
        )

    def get_service(self, name: str, **kwargs: Any) -> Response[ServiceItem]:
        self.__setup__()
        app_runner_service = self._get_service_by_name(name)
        if not app_runner_service:
            raise NotFoundError(f"Service {name} not found.")
        try:
            response = self._client.describe_service(
                ServiceArn=app_runner_service["ServiceArn"]
            )
            result = self._result_converter.convert_service_item(
                response["Service"]
            )
            return Response(result=result)
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                raise NotFoundError(f"Service {name} not found.")
            raise

    def delete_service(self, name: str, **kwargs: Any) -> Response[None]:
        self.__setup__()
        app_runner_service = self._get_service_by_name(name)
        if not app_runner_service:
            raise NotFoundError(f"Service {name} not found.")
        try:
            _ = self._client.delete_service(
                ServiceArn=app_runner_service["ServiceArn"]
            )
            # Wait for deletion to complete
            self._wait_for_deletion(app_runner_service["ServiceArn"])
            if not self.auto_scaling_configuration_arn:
                self._delete_auto_scaling_configuration(name)
            print(f"Service {name} deleted successfully.")
            return Response(result=None)
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                raise NotFoundError(f"Service {name} not found.")
            raise

    def list_services(self, **kwargs: Any) -> Response[list[ServiceItem]]:
        self.__setup__()
        response = self._client.list_services()
        services = []

        for service_summary in response.get("ServiceSummaryList", []):
            # Get detailed service information
            detailed_response = self._client.describe_service(
                ServiceArn=service_summary["ServiceArn"]
            )
            services.append(detailed_response["Service"])

        result = self._result_converter.convert_services(services)
        return Response(result=result)

    def list_revisions(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[list[Revision]]:
        self.__setup__()
        app_runner_service = self._get_service_by_name(name)
        if not app_runner_service:
            raise NotFoundError(f"Service {name} not found.")
        try:
            response = self._client.describe_service(
                ServiceArn=app_runner_service["ServiceArn"]
            )
            result = self._result_converter.convert_service_item(
                response["Service"]
            )
            revision = Revision(
                name="latest",
                traffic=100,
                containers=result.service.containers if result.service else [],
                volumes=result.service.volumes if result.service else [],
                created_time=None,
                active=True,
                status="running",
            )
            return Response(result=[revision])
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                raise NotFoundError(f"Service {name} not found.")
            raise

    def get_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[Revision]:
        if revision != "latest":
            raise NotFoundError("Only 'latest' revision is supported.")
        self.__setup__()
        app_runner_service = self._get_service_by_name(name)
        if not app_runner_service:
            raise NotFoundError(f"Service {name} not found.")
        try:
            response = self._client.describe_service(
                ServiceArn=app_runner_service["ServiceArn"]
            )
            result = self._result_converter.convert_service_item(
                response["Service"]
            )
            return Response(
                result=Revision(
                    name="latest",
                    traffic=100,
                    containers=(
                        result.service.containers if result.service else []
                    ),
                    volumes=result.service.volumes if result.service else [],
                    created_time=None,
                    active=True,
                    status="running",
                )
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                raise NotFoundError(f"Service {name} not found.")
            raise

    def delete_revision(self, name: str, revision: str) -> Response[None]:
        self.__setup__()
        app_runner_service = self._get_service_by_name(name)
        if not app_runner_service:
            raise NotFoundError(f"Service {name} not found.")
        if revision != "latest":
            raise NotFoundError("Only 'latest' revision is supported.")
        try:
            self._client.describe_service(
                ServiceArn=app_runner_service["ServiceArn"]
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                raise NotFoundError(f"Service {name} not found.")
            raise
        raise PreconditionFailedError(
            (
                f"Cannot delete active revision {revision} "
                f"of service {name}."
            )
        )

    def update_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
    ) -> Response[ServiceItem]:
        return self.get_service(name)

    def close(self) -> None:
        self._init = False

    def _get_service_name(self, service: ServiceDefinition) -> str:
        if self.service_name:
            return self.service_name
        if service.name:
            return service.name
        raise BadRequestError(
            "Service name must be provided or defined in the service."
        )

    def _get_service_by_name(self, name: str) -> dict[str, Any] | None:
        response = self._client.list_services()
        services = response.get("ServiceSummaryList", [])
        for service in services:
            if service["ServiceName"] == name:
                return service
        return None

    def _wait_for_creation(self, operation_id: str, service_arn: str) -> None:
        max_attempts = 60
        delay = 5  # seconds
        for attempt in range(max_attempts):
            response = self._client.describe_service(ServiceArn=service_arn)
            status = response["Service"]["Status"]
            if status in ("RUNNING", "PAUSED"):
                return
            elif status in ("FAILED", "DELETED"):
                raise RuntimeError(
                    f"Service operation failed with status: {status}"
                )
            time.sleep(delay)

    def _wait_for_deletion(self, service_arn: str) -> None:
        max_attempts = 60
        delay = 5  # seconds

        for attempt in range(max_attempts):
            try:
                response = self._client.describe_service(
                    ServiceArn=service_arn
                )
                status = response["Service"]["Status"]
                if status == "DELETED":
                    return
            except ClientError as e:
                if e.response["Error"]["Code"] == "ResourceNotFoundException":
                    return
                raise

            time.sleep(delay)

    def _get_or_create_ecr_connection(self) -> str:
        connection_name = "default-ecr-connection"
        arn = self._get_ecr_connection_arn()
        if arn:
            return arn

        try:
            response = self._client.create_connection(
                ConnectionName=connection_name,
                ProviderType="ECR",
            )
            connection_arn = response["Connection"]["ConnectionArn"]
            return connection_arn
        except ClientError as e:
            raise RuntimeError(f"Failed to create connection: {e}")

    def _get_ecr_connection_arn(self) -> str | None:
        connections = self._client.list_connections()
        for conn in connections["ConnectionSummaryList"]:
            if conn["ProviderType"] == "ECR":
                return conn["ConnectionArn"]
        return None

    def _get_or_create_auto_scaling_configuration(
        self,
        service_name: str,
        scale: Scale | None = None,
    ) -> str:
        max_concurrency = (
            scale.max_concurrency if scale and scale.max_concurrency else 100
        )
        min_size = scale.min_replicas if scale and scale.min_replicas else 1
        max_size = scale.max_replicas if scale and scale.max_replicas else 1
        existing_arn = self._get_active_auto_scaling_config_arn(
            service_name, max_concurrency, min_size, max_size
        )
        if existing_arn:
            return existing_arn

        response = self._client.create_auto_scaling_configuration(
            AutoScalingConfigurationName=service_name,
            MaxConcurrency=max_concurrency,
            MinSize=min_size,
            MaxSize=max_size,
        )
        arn = response["AutoScalingConfiguration"][
            "AutoScalingConfigurationArn"
        ]
        return arn

    def _get_active_auto_scaling_config_arn(
        self,
        name: str,
        max_concurrency: int,
        min_size: int,
        max_size: int,
    ) -> str | None:
        configs = self._client.list_auto_scaling_configurations()
        for config in configs["AutoScalingConfigurationSummaryList"]:
            if config["AutoScalingConfigurationName"] == name:
                full_config = self._client.describe_auto_scaling_configuration(
                    AutoScalingConfigurationArn=config[
                        "AutoScalingConfigurationArn"
                    ]
                )["AutoScalingConfiguration"]
                if (
                    full_config["MaxConcurrency"] == max_concurrency
                    and full_config["MinSize"] == min_size
                    and full_config["MaxSize"] == max_size
                ):
                    print("Found existing auto scaling configuration")
                    return config["AutoScalingConfigurationArn"]
        return None

    def _delete_auto_scaling_configuration(self, name: str) -> None:
        configs = self._client.list_auto_scaling_configurations()
        for config in configs["AutoScalingConfigurationSummaryList"]:
            if config["AutoScalingConfigurationName"] == name:
                self._client.delete_auto_scaling_configuration(
                    AutoScalingConfigurationArn=config[
                        "AutoScalingConfigurationArn"
                    ]
                )

    def _create_ecr_access_role(self) -> str:
        trust_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "build.apprunner.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        role_name = "AppRunnerECRAccessRole"
        try:
            role_response = self._iam_client.create_role(
                RoleName="AppRunnerECRAccessRole",
                AssumeRolePolicyDocument=json.dumps(trust_policy),
                Description="Role for App Runner to access ECR",
            )
            role_arn = role_response["Role"]["Arn"]
        except self._iam_client.exceptions.EntityAlreadyExistsException:
            # Role already exists, get its ARN
            role_response = self._iam_client.get_role(RoleName=role_name)
            role_arn = role_response["Role"]["Arn"]
        self._iam_client.attach_role_policy(
            RoleName=role_name,
            PolicyArn="arn:aws:iam::aws:policy/service-role/AWSAppRunnerServicePolicyForECRAccess",  # noqa
        )
        return role_arn


class OperationConverter:
    def convert_service_create(
        self,
        service_name: str,
        service: ServiceDefinition,
        images: list[str],
        ecr_connection_arn: str | None = None,
        auto_scaling_config_arn: str | None = None,
        vpc_connector_arn: str | None = None,
        authentication_configuration_access_role_arn: str | None = None,
    ) -> dict[str, Any]:
        if len(service.containers) != 1:
            raise BadRequestError(
                "App Runner supports only one container per service"
            )

        container = service.containers[0]
        image = images[0]
        config = {
            "ServiceName": service_name,
            "SourceConfiguration": self._convert_source_configuration(
                container,
                image,
                ecr_connection_arn,
                authentication_configuration_access_role_arn,
            ),
            "InstanceConfiguration": self._convert_instance_configuration(
                container.resources,
            ),
        }

        config["NetworkConfiguration"] = self._convert_network_configuration(
            ingress=service.ingress,
            vpc_connector_arn=vpc_connector_arn,
        )

        if container.probes and container.probes.readiness_probe:
            config["HealthCheckConfiguration"] = (
                self._convert_health_check_configuration(
                    container.probes.readiness_probe
                )
            )

        if auto_scaling_config_arn:
            config["AutoScalingConfigurationArn"] = auto_scaling_config_arn

        return config

    def convert_service_update(
        self,
        service_arn: str,
        service: ServiceDefinition,
        images: list[str],
        ecr_connection_arn: str | None = None,
        auto_scaling_config_arn: str | None = None,
        vpc_connector_arn: str | None = None,
    ) -> dict[str, Any]:
        if len(service.containers) != 1:
            raise BadRequestError(
                "App Runner supports only one container per service"
            )

        container = service.containers[0]
        image = images[0]

        config = {
            "ServiceArn": service_arn,
            "SourceConfiguration": self._convert_source_configuration(
                container, image, ecr_connection_arn
            ),
            "NetworkConfiguration": self._convert_network_configuration(
                service.ingress, vpc_connector_arn
            ),
        }
        if auto_scaling_config_arn:
            config["AutoScalingConfigurationArn"] = auto_scaling_config_arn
        return config

    def _convert_source_configuration(
        self,
        container: Container,
        image: str,
        ecr_connection_arn: str | None = None,
        authentication_configuration_access_role_arn: str | None = None,
    ) -> dict[str, Any]:
        image_identifier = image
        if ":" not in image:
            image_identifier = f"{image}:latest"
        config: dict = {
            "ImageRepository": {
                "ImageRepositoryType": "ECR",
                "ImageIdentifier": image_identifier,
                "ImageConfiguration": self._convert_image_configuration(
                    container
                ),
            },
            "AutoDeploymentsEnabled": False,
        }
        if ecr_connection_arn:
            config["ImageRepository"][
                "ImageRepositoryConnectionArn"
            ] = ecr_connection_arn
        if authentication_configuration_access_role_arn:
            config["AuthenticationConfiguration"] = {
                "AccessRoleArn": authentication_configuration_access_role_arn
            }
        return config

    def _convert_image_configuration(
        self,
        container: Container,
    ) -> dict[str, Any]:
        config: dict[str, Any] = {}
        if container.ports:
            config["Port"] = str(container.ports[0].container_port)
        if container.command:
            if container.args:
                full_command = container.command + container.args
            else:
                full_command = container.command
            config["StartCommand"] = " ".join(full_command)
        if container.env:
            config["RuntimeEnvironmentVariables"] = {
                var.name: str(var.value) for var in container.env
            }
        return config

    def _convert_instance_configuration(
        self,
        resources: ResourceRequirements | None,
    ) -> dict[str, Any]:
        config: dict[str, Any] = {}
        if resources and resources.requests and resources.requests.cpu:
            cpu = f"{resources.requests.cpu} vCPU"
        else:
            cpu = "0.25 vCPU"

        if resources and resources.requests and resources.requests.memory:
            # convert memory mb to gb
            memory_gb = resources.requests.memory / 1024
        else:
            memory_gb = 0.5

        if memory_gb <= 0.5:
            memory = "0.5 GB"
        elif memory_gb <= 1:
            memory = "1 GB"
        elif memory_gb <= 2:
            memory = "2 GB"
        elif memory_gb <= 3:
            memory = "3 GB"
        elif memory_gb <= 4:
            memory = "4 GB"
        else:
            memory = "8 GB"

        config["Cpu"] = cpu
        config["Memory"] = memory
        return config

    def _convert_health_check_configuration(
        self,
        probe: Probe,
    ) -> dict[str, Any]:
        config: dict[str, Any] = {}

        if probe.http_get:
            config["Protocol"] = "HTTP"
            config["Path"] = probe.http_get.path or "/"
        elif probe.tcp_socket:
            config["Protocol"] = "TCP"
        else:
            config["Protocol"] = "HTTP"
            config["Path"] = "/"

        if probe.period_seconds is not None:
            config["Interval"] = probe.period_seconds
        if probe.timeout_seconds is not None:
            config["Timeout"] = probe.timeout_seconds
        if probe.success_threshold is not None:
            config["HealthyThreshold"] = probe.success_threshold
        if probe.failure_threshold is not None:
            config["UnhealthyThreshold"] = probe.failure_threshold

        return config

    def _convert_network_configuration(
        self,
        ingress: Ingress | None = None,
        vpc_connector_arn: str | None = None,
    ) -> dict[str, Any]:
        config: dict[str, Any] = {}
        if not ingress or ingress.external:
            config["EgressConfiguration"] = {"EgressType": "DEFAULT"}
        else:
            config["EgressConfiguration"] = {
                "EgressType": "VPC",
                "VpcConnectorArn": vpc_connector_arn,
            }
        return config


class ResultConverter:
    def convert_service_item(self, service: dict[str, Any]) -> ServiceItem:
        return ServiceItem(
            name=service["ServiceName"],
            uri=f"https://{service.get("ServiceUrl", "")}",
            service=self.convert_service(service),
        )

    def convert_service(self, service: dict[str, Any]) -> ServiceDefinition:
        # Extract container info from source configuration
        source_config = service.get("SourceConfiguration", {})
        image_repo = source_config.get("ImageRepository", {})
        image_config = image_repo.get("ImageConfiguration", {})
        heath_check_config = service.get("HealthCheckConfiguration", {})
        network_config = service.get("NetworkConfiguration", {})

        # Create container from service configuration
        container = Container(
            name=image_repo.get("ImageIdentifier", "")
            .split("/")[-1]
            .split(":")[0],
            type="main",
            image=image_repo.get("ImageIdentifier", ""),
            env=self._convert_env(
                image_config.get("RuntimeEnvironmentVariables", {})
            ),
            ports=self._convert_ports(image_config.get("Port", None)),
            resources=self._convert_resources(
                service.get("InstanceConfiguration", {})
            ),
            probes=self._convert_probes(
                heath_check_config, image_config.get("Port", None)
            ),
        )

        return ServiceDefinition(
            name=service["ServiceName"],
            containers=[container],
            ingress=self._convert_ingress(network_config),
            latest_ready_revision="latest",
            latest_created_revision="latest",
            traffic=[TrafficAllocation(revision="latest", percent=100)],
        )

    def _convert_ingress(
        self, network_config: dict[str, Any]
    ) -> Ingress | None:
        egress_config = network_config.get("EgressConfiguration", {})
        if egress_config.get("EgressType") == "VPC":
            return Ingress(
                external=False,
            )
        return Ingress(external=True)

    def _convert_probes(
        self,
        health_check_config: dict[str, Any],
        port: str | None = None,
    ) -> ProbeSet | None:
        if not health_check_config:
            return None

        readiness_probe = Probe(
            period_seconds=health_check_config.get("Interval", None),
            timeout_seconds=health_check_config.get("Timeout", None),
            success_threshold=health_check_config.get(
                "HealthyThreshold", None
            ),
            failure_threshold=health_check_config.get(
                "UnhealthyThreshold", None
            ),
        )
        if health_check_config.get("Protocol") == "HTTP":
            readiness_probe.http_get = HTTPGetAction(
                path=health_check_config.get("Path", "/"),
            )
        elif health_check_config.get("Protocol") == "TCP":
            readiness_probe.tcp_socket = TCPSocketAction(
                port=health_check_config.get("Port", port or "80")
            )

        probes = ProbeSet(readiness_probe=readiness_probe)
        return probes

    def _convert_env(self, env_vars: dict[str, str]) -> list[EnvVar]:
        return [
            EnvVar(name=name, value=value) for name, value in env_vars.items()
        ]

    def _convert_ports(self, port: str | None) -> list:
        if not port:
            return []
        return [Port(container_port=int(port), protocol="TCP")]

    def _convert_resources(
        self, instance_config: dict[str, Any]
    ) -> ResourceRequirements | None:
        if not instance_config:
            return None

        cpu_str = instance_config.get("Cpu", "0.25 vCPU")
        memory_str = instance_config.get("Memory", "0.5 GB")

        cpu = None
        memory = None
        if cpu_str.endswith(" vCPU"):
            cpu_value = cpu_str[:-5]
            cpu = float(cpu_value)

        if memory_str.endswith(" GB"):
            memory_value = float(memory_str[:-3])
            memory = int(memory_value * 1024)

        return ResourceRequirements(
            requests=ResourceRequests(cpu=cpu, memory=memory)
        )

    def convert_services(
        self, services: list[dict[str, Any]]
    ) -> list[ServiceItem]:
        return [self.convert_service_item(service) for service in services]



from ._attributes import (
    Attribute,
    KeyAttribute,
    SpecialAttribute,
    UpdateAttribute,
)
from ._comparator import Comparator
from ._component import StoreComponent
from ._index_helper import IndexHelper
from ._item_processor import ItemProcessor
from ._models import (
    ArrayIndex,
    AscIndex,
    BaseIndex,
    CollectionResult,
    CollectionStatus,
    CompositeIndex,
    DescIndex,
    ExcludeIndex,
    FieldIndex,
    GeospatialFieldType,
    GeospatialIndex,
    HashIndex,
    Index,
    IndexResult,
    IndexStatus,
    IndexType,
    MatchCondition,
    RangeIndex,
    RankIndex,
    RankMethod,
    SparseVectorIndex,
    TextIndex,
    TextSimilarityAlgorithm,
    TTLIndex,
    VectorIndex,
    VectorIndexMetric,
    VectorIndexStructure,
    WildcardIndex,
)
from ._operation import StoreOperation
from ._operation_parser import StoreOperationParser
from ._parameter_parser import ParameterParser
from ._provider import StoreProvider
from ._validator import Validator

__all__ = [
    "Attribute",
    "Comparator",
    "ParameterParser",
    "IndexHelper",
    "ItemProcessor",
    "KeyAttribute",
    "MatchCondition",
    "SpecialAttribute",
    "StoreComponent",
    "StoreOperation",
    "StoreOperationParser",
    "StoreProvider",
    "UpdateAttribute",
    "Validator",
    "MatchCondition",
    "ArrayIndex",
    "AscIndex",
    "BaseIndex",
    "CompositeIndex",
    "DescIndex",
    "ExcludeIndex",
    "FieldIndex",
    "GeospatialFieldType",
    "GeospatialIndex",
    "HashIndex",
    "Index",
    "IndexType",
    "RangeIndex",
    "RankIndex",
    "RankMethod",
    "SparseVectorIndex",
    "TextIndex",
    "TextSimilarityAlgorithm",
    "TTLIndex",
    "VectorIndex",
    "VectorIndexMetric",
    "VectorIndexStructure",
    "WildcardIndex",
    "IndexStatus",
    "IndexResult",
    "CollectionStatus",
    "CollectionResult",
]



from x8.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import QueryFunction
from x8.storage._common import CollectionResult, CollectionStatus

from ._models import (
    VectorBatch,
    VectorCollectionConfig,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorTransaction,
    VectorValue,
)
from .component import VectorStore

__all__ = [
    "VectorBatch",
    "VectorCollectionConfig",
    "VectorItem",
    "VectorKey",
    "VectorList",
    "VectorProperties",
    "VectorStore",
    "VectorTransaction",
    "VectorValue",
    "QueryFunction",
    "BadRequestError",
    "ConflictError",
    "NotFoundError",
    "PreconditionFailedError",
    "CollectionResult",
    "CollectionStatus",
]



from typing import Any

from x8.core import Context, Operation, Provider, warn
from x8.core.exceptions import BadRequestError

from .._models import ChoiceProviderInfo


class Default(Provider):
    providers: list[ChoiceProviderInfo]

    _provider_map: dict[str, Provider]

    def __init__(self, providers: list[ChoiceProviderInfo], **kwargs):
        self.providers = providers
        self._init_provider_map()

    def _init_provider_map(self):
        self._provider_map = {
            provider.key: provider.provider for provider in self.providers
        }

    def _get_choice_provider(self, context: Context | None = None) -> Provider:
        if (
            context
            and context.data
            and "provider" in context.data
            and context.data["provider"] in self._provider_map
        ):
            return self._provider_map[context.data["provider"]]
        warn("Choice provider not found. Falling back to the first provider")
        if not self.providers:
            raise BadRequestError("No choice providers available")

        return self.providers[0].provider

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        provider = self._get_choice_provider(context=context)
        return provider.__run__(operation=operation, context=context, **kwargs)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        provider = self._get_choice_provider(context=context)
        return await provider.__arun__(
            operation=operation, context=context, **kwargs
        )



import os
import time

import httpx
import pytest

from x8.compute.container_deployment import (
    Container,
    ContainerDeploymentFeature,
    EnvVar,
    ImageMap,
    Port,
    Revision,
    ServiceDefinition,
    ServiceItem,
    TrafficAllocation,
)
from x8.compute.containerizer import BuildConfig
from x8.core import Loader, RunContext
from x8.core.exceptions import NotFoundError, PreconditionFailedError

from ._providers import ContainerDeploymentProvider
from ._sync_and_async_client import ContainerDeploymentClient


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        ContainerDeploymentProvider.AWS_APP_RUNNER,
        ContainerDeploymentProvider.AMAZON_ECS_EC2,
        ContainerDeploymentProvider.AMAZON_ECS_FARGATE,
        ContainerDeploymentProvider.AZURE_CONTAINER_INSTANCES,
        ContainerDeploymentProvider.AZURE_CONTAINER_APPS,
        ContainerDeploymentProvider.GOOGLE_CLOUD_RUN,
        ContainerDeploymentProvider.LOCAL,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_single_container(provider_type: str, async_call: bool):
    client = ContainerDeploymentClient(
        provider_type=provider_type, async_call=async_call
    )
    service_name = "X8test2"
    try:
        await client.delete_service(name=service_name)
        wait_to_delete(provider_type=provider_type)
    except NotFoundError:
        pass

    with pytest.raises(NotFoundError):
        await client.get_service(name=service_name)

    base_path = os.path.dirname(os.path.abspath(__file__))
    service_path = os.path.join(base_path, "service")
    loader = Loader(path=service_path)
    api_main = loader.load_component("api_single_main")
    service_def = get_single_container_service_definition(
        service_name, api_main, "v1"
    )
    res = await client.create_service(
        service=service_def,
        run_context=RunContext(path=service_path),
    )
    response = httpx.post(
        url=f"{res.result.uri}/add",
        json={"a": "5", "b": "10"},
    )
    assert response.json() == 15
    response = httpx.post(
        url=f"{res.result.uri}/version",
    )
    assert response.json() == "v1"

    res = await client.get_service(name=service_name)
    assert_service_definition(res.result, service_def)
    revision_name = res.result.service.latest_created_revision
    assert_traffic(
        res.result.service.traffic,
        [
            TrafficAllocation(
                revision=revision_name,
                percent=100,
            )
        ],
    )

    res = await client.list_services()
    assert service_name in [s.name for s in res.result]

    res = await client.list_revisions(name=service_name)
    assert_revisions(res.result, [Revision(name=revision_name, traffic=100)])

    res = await client.get_revision(
        name=service_name,
        revision=res.result[0].name,
    )
    assert_revision(res.result, service_def)

    with pytest.raises(PreconditionFailedError):
        await client.delete_revision(
            name=service_name,
            revision=res.result.name,
        )

    service_def_2 = get_single_container_service_definition(
        service_name, api_main, "v2"
    )
    with pytest.raises(PreconditionFailedError):
        res = await client.create_service(
            service=service_def_2,
            run_context=RunContext(path=service_path),
            where="not_exists()",
        )

    res = await client.create_service(
        service=service_def_2,
        run_context=RunContext(path=service_path),
    )
    response = httpx.post(
        url=f"{res.result.uri}/subtract",
        json={"a": "5", "b": "10"},
    )
    assert response.json() == -5
    response = httpx.post(
        url=f"{res.result.uri}/version",
    )
    assert response.json() == "v2"

    res = await client.get_service(name=service_name)
    assert_service_definition(res.result, service_def_2)

    if client.client.__supports__(
        ContainerDeploymentFeature.MULTIPLE_REVISIONS
    ):
        res = await client.list_revisions(name=service_name)
        revisions = res.result
        current_revision = res.result[0].name
        old_revision = res.result[1].name
        for rev in res.result:
            if rev.traffic == 100:
                current_revision = rev.name
            else:
                old_revision = rev.name
        assert_revisions(
            res.result,
            [
                Revision(name=current_revision, traffic=100),
                Revision(name=old_revision, traffic=0),
            ],
        )

        res = await client.get_revision(
            name=service_name,
            revision=current_revision,
        )
        assert_revision(res.result, service_def_2)

        traffic = []
        for rev in revisions:
            if rev.traffic == 100:
                traffic.append(TrafficAllocation(revision=rev.name, percent=0))
            else:
                traffic.append(
                    TrafficAllocation(revision=rev.name, percent=100)
                )
        res = await client.update_traffic(
            name=service_name,
            traffic=traffic,
        )
        response = httpx.post(
            url=f"{res.result.uri}/do",
            json={"a": "5"},
        )
        assert response.json() == 15

        if client.client.__supports__(
            ContainerDeploymentFeature.REVISION_DELETE
        ):
            # Go back to the old traffic allocation
            # We cannot delete revisions that have traffic
            # allocated or is the latest revision
            traffic = []
            for rev in revisions:
                if rev.traffic == 0:
                    traffic.append(
                        TrafficAllocation(revision=rev.name, percent=0)
                    )
                else:
                    traffic.append(
                        TrafficAllocation(revision=rev.name, percent=100)
                    )

            res = await client.update_traffic(
                name=service_name,
                traffic=traffic,
            )
            await client.delete_revision(
                name=service_name,
                revision=old_revision,
            )

            res = await client.list_revisions(name=service_name)
            assert_revisions(
                res.result,
                [
                    Revision(name=current_revision, traffic=100),
                ],
            )

    await client.delete_service(name=service_name)
    wait_to_delete(provider_type=provider_type)

    with pytest.raises(NotFoundError):
        await client.get_service(name=service_name)

    # container_registry = client.client.__provider__.container_registry
    # for container in service_def.containers:
    #    image_name = container.name
    #    container_registry.delete(image_name=image_name)

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        ContainerDeploymentProvider.AMAZON_ECS_FARGATE,
        ContainerDeploymentProvider.AMAZON_ECS_EC2,
        ContainerDeploymentProvider.AZURE_CONTAINER_INSTANCES,
        ContainerDeploymentProvider.AZURE_CONTAINER_APPS,
        ContainerDeploymentProvider.GOOGLE_CLOUD_RUN,
        ContainerDeploymentProvider.LOCAL,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_multiple_containers(provider_type: str, async_call: bool):
    client = ContainerDeploymentClient(
        provider_type=provider_type, async_call=async_call
    )
    service_name = f"X8test1-{str(async_call).lower()}"
    try:
        await client.delete_service(name=service_name)
        wait_to_delete(provider_type=provider_type)
    except NotFoundError:
        pass

    with pytest.raises(NotFoundError):
        await client.get_service(name=service_name)

    base_path = os.path.dirname(os.path.abspath(__file__))
    service_path = os.path.join(base_path, "service")
    loader = Loader(path=service_path)
    api_main = loader.load_component("api_main")
    api_side = loader.load_component("api_side")
    service_def = get_multi_container_service_definition(
        service_name, api_main, api_side, "v1"
    )
    res = await client.create_service(
        service=service_def,
        run_context=RunContext(path=service_path, tag="add"),
    )
    response = httpx.post(
        url=f"{res.result.uri}/do",
        json={"a": "5"},
    )
    assert response.json() == 15

    res = await client.get_service(name=service_name)
    assert_service_definition(res.result, service_def)
    revision_name = res.result.service.latest_created_revision
    assert_traffic(
        res.result.service.traffic,
        [
            TrafficAllocation(
                revision=revision_name,
                percent=100,
            )
        ],
    )

    res = await client.list_services()
    assert service_name in [s.name for s in res.result]

    res = await client.list_revisions(name=service_name)
    assert_revisions(res.result, [Revision(name=revision_name, traffic=100)])

    res = await client.get_revision(
        name=service_name,
        revision=res.result[0].name,
    )
    assert_revision(res.result, service_def)

    with pytest.raises(PreconditionFailedError):
        await client.delete_revision(
            name=service_name,
            revision=res.result.name,
        )

    service_def_2 = get_multi_container_service_definition(
        service_name, api_main, api_side, "v2"
    )
    with pytest.raises(PreconditionFailedError):
        res = await client.create_service(
            service=service_def_2,
            run_context=RunContext(path=service_path, tag="subtract"),
            where="not_exists()",
        )

    res = await client.create_service(
        service=service_def_2,
        run_context=RunContext(path=service_path, tag="subtract"),
    )
    time.sleep(5)
    response = httpx.post(
        url=f"{res.result.uri}/do",
        json={"a": "5"},
    )
    assert response.json() == -5

    res = await client.get_service(name=service_name)
    assert_service_definition(res.result, service_def_2)
    revision_name = res.result.service.latest_created_revision

    if client.client.__supports__(
        ContainerDeploymentFeature.MULTIPLE_REVISIONS
    ):
        res = await client.list_revisions(name=service_name)
        revisions = res.result
        current_revision = res.result[0].name
        old_revision = res.result[1].name
        for rev in res.result:
            if rev.traffic == 100:
                current_revision = rev.name
            else:
                old_revision = rev.name
        assert_revisions(
            res.result,
            [
                Revision(name=current_revision, traffic=100),
                Revision(name=old_revision, traffic=0),
            ],
        )

        res = await client.get_revision(
            name=service_name,
            revision=current_revision,
        )
        assert_revision(res.result, service_def_2)

        traffic = []
        for rev in revisions:
            if rev.traffic == 100:
                traffic.append(TrafficAllocation(revision=rev.name, percent=0))
            else:
                traffic.append(
                    TrafficAllocation(revision=rev.name, percent=100)
                )
        res = await client.update_traffic(
            name=service_name,
            traffic=traffic,
        )
        response = httpx.post(
            url=f"{res.result.uri}/do",
            json={"a": "5"},
        )
        assert response.json() == 15

        if client.client.__supports__(
            ContainerDeploymentFeature.REVISION_DELETE
        ):
            # Go back to the old traffic allocation
            # We cannot delete revisions that have traffic
            # allocated or is the latest revision
            traffic = []
            for rev in revisions:
                if rev.traffic == 0:
                    traffic.append(
                        TrafficAllocation(revision=rev.name, percent=0)
                    )
                else:
                    traffic.append(
                        TrafficAllocation(revision=rev.name, percent=100)
                    )

            res = await client.update_traffic(
                name=service_name,
                traffic=traffic,
            )
            await client.delete_revision(
                name=service_name,
                revision=old_revision,
            )

            res = await client.list_revisions(name=service_name)
            assert_revisions(
                res.result,
                [
                    Revision(name=current_revision, traffic=100),
                ],
            )
    else:
        res = await client.list_revisions(name=service_name)
        assert_revisions(
            res.result, [Revision(name=revision_name, traffic=100)]
        )

        res = await client.get_revision(
            name=service_name,
            revision=res.result[0].name,
        )

    await client.delete_service(name=service_name)
    wait_to_delete(provider_type=provider_type)

    with pytest.raises(NotFoundError):
        await client.get_service(name=service_name)

    # container_registry = client.client.__provider__.container_registry
    # for container in service_def.containers:
    #    image_name = container.name
    #    container_registry.delete(image_name=image_name)

    await client.close()


def get_single_container_service_definition(
    name, main, version
) -> ServiceDefinition:
    return ServiceDefinition(
        name=name,
        images=[
            ImageMap(
                name="main",
                component=main,
            ),
        ],
        containers=[
            Container(
                name="main",
                ports=[
                    Port(container_port=80),
                ],
                env=[
                    EnvVar(name="BACKEND_URL", value="http://localhost:8000"),
                    EnvVar(name="VERSION", value=version),
                ],
            ),
        ],
    )


def get_multi_container_service_definition(
    name, main, side, version
) -> ServiceDefinition:
    return ServiceDefinition(
        name=name,
        images=[
            ImageMap(
                name="main",
                component=main,
                build=BuildConfig(
                    image_name=f"main:{version}",
                    nocache=False,
                ),
            ),
            ImageMap(
                name="side",
                component=side,
                build=BuildConfig(
                    image_name=f"side:{version}",
                    nocache=False,
                ),
            ),
        ],
        containers=[
            Container(
                name="main",
                ports=[
                    Port(container_port=80),
                ],
                env=[
                    EnvVar(name="BACKEND_URL", value="http://localhost:8000"),
                    EnvVar(name="VERSION", value=version),
                ],
            ),
            Container(name="side"),
        ],
    )


def assert_service_definition(
    result: ServiceItem, expected: ServiceDefinition
):
    assert result.name == expected.name
    assert result.uri is not None
    assert result.service is not None
    assert result.service.name == expected.name
    assert len(result.service.containers) == len(expected.containers)
    for container in result.service.containers:
        assert_container(
            result=container,
            expected=next(
                c for c in expected.containers if c.name == container.name
            ),
        )


def assert_revision(result: Revision, expected: ServiceDefinition):
    for container in result.containers:
        assert_container(
            result=container,
            expected=next(
                c for c in expected.containers if c.name == container.name
            ),
        )


def assert_revisions(result: list[Revision], expected: list[Revision]):
    assert len(result) == len(expected)
    for item in result:
        assert item.name in [e.name for e in expected]
        assert item.traffic in [e.traffic for e in expected]


def assert_traffic(
    result: list[TrafficAllocation],
    expected: list[TrafficAllocation],
):
    assert len(result) == len(expected)
    for item in result:
        assert item.revision is not None
        assert item.percent in [e.percent for e in expected]


def assert_container(result: Container, expected: Container):
    assert result.name == expected.name
    # assert len(result.ports) == len(expected.ports)
    # assert_ports(result.ports, expected.ports)
    assert_env(result.env, expected.env)


def assert_ports(result: list[Port], expected: list[Port]):
    assert len(result) == len(expected)
    for port in result:
        assert port.container_port in [p.container_port for p in expected]


def assert_env(result: list[EnvVar], expected: list[EnvVar]):
    assert len(result) == len(expected)
    for env in result:
        assert env.name in [e.name for e in expected]
        assert env.value == next(
            e.value for e in expected if e.name == env.name
        )


def wait_to_delete(provider_type: str):
    if provider_type == ContainerDeploymentProvider.GOOGLE_CLOUD_RUN:
        time.sleep(10)
